# Copyright 2021-2022 python-tuf contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Repository Abstraction for metadata management"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager, suppress
from copy import deepcopy
from typing import TYPE_CHECKING

from tuf.api.exceptions import UnsignedMetadataError
from tuf.api.metadata import (
    Metadata,
    MetaFile,
    Root,
    Signed,
    Snapshot,
    Targets,
    Timestamp,
)

if TYPE_CHECKING:
    from collections.abc import Generator

logger = logging.getLogger(__name__)


class AbortEdit(Exception):  # noqa: N818
    """Raise to exit the edit() contextmanager without saving changes"""


class Repository(ABC):
    """Abstract class for metadata modifying implementations

    NOTE: The repository module is not considered part of the python-tuf
    stable API yet.

    This class is intended to be a base class used in any metadata editing
    application, whether it is a real repository server or a developer tool.

    Implementations must implement open() and close(), and can then use the
    edit() contextmanager to implement actual operations. Note that signing
    an already existing version of metadata (as could be done for threshold
    signing) does not fit into this model of open()+close() or edit().

    A few operations (snapshot and timestamp) are already implemented
    in this base class.
    """

    @abstractmethod
    def open(self, role: str) -> Metadata:
        """Load a roles metadata from storage or cache, return it

        If role has no metadata, create first version from scratch.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self, role: str, md: Metadata) -> None:
        """Write roles metadata into storage

        Update expiry and version and replace signatures with ones from all
        available keys. Keep snapshot_info and targets_infos updated.
        """
        raise NotImplementedError

    @property
    def targets_infos(self) -> dict[str, MetaFile]:
        """Returns the MetaFiles for current targets metadatas

        This property is used by do_snapshot() to update Snapshot.meta:
        Repository implementations should override this property to enable
        do_snapshot().

        Note that there is a difference between this return value and
        Snapshot.meta: This dictionary reflects the targets metadata that
        currently exists in the repository but Snapshot.meta also includes
        metadata that used to exist, but no longer exists, in the repository.
        """
        raise NotImplementedError

    @property
    def snapshot_info(self) -> MetaFile:
        """Returns the MetaFile for current snapshot metadata

        This property is used by do_timestamp() to update Timestamp.meta:
        Repository implementations should override this property to enable
        do_timestamp().
        """
        raise NotImplementedError

    @contextmanager
    def edit(self, role: str) -> Generator[Signed, None, None]:
        """Context manager for editing a role's metadata

        Context manager takes care of loading the roles metadata (or creating
        new metadata), updating expiry and version. The caller can do
        other changes to the Signed object and when the context manager exits,
        a new version of the roles metadata is stored.

        Context manager user can raise AbortEdit from inside the with-block to
        cancel the edit: in this case none of the changes are stored.
        """
        md = self.open(role)
        with suppress(AbortEdit):
            yield md.signed
            self.close(role, md)

    @contextmanager
    def edit_root(self) -> Generator[Root, None, None]:
        """Context manager for editing root metadata. See edit()"""
        with self.edit(Root.type) as root:
            if not isinstance(root, Root):
                raise AssertionError("Unexpected root type")
            yield root

    @contextmanager
    def edit_timestamp(self) -> Generator[Timestamp, None, None]:
        """Context manager for editing timestamp metadata. See edit()"""
        with self.edit(Timestamp.type) as timestamp:
            if not isinstance(timestamp, Timestamp):
                raise AssertionError("Unexpected timestamp type")
            yield timestamp

    @contextmanager
    def edit_snapshot(self) -> Generator[Snapshot, None, None]:
        """Context manager for editing snapshot metadata. See edit()"""
        with self.edit(Snapshot.type) as snapshot:
            if not isinstance(snapshot, Snapshot):
                raise AssertionError("Unexpected snapshot type")
            yield snapshot

    @contextmanager
    def edit_targets(
        self, rolename: str = Targets.type
    ) -> Generator[Targets, None, None]:
        """Context manager for editing targets metadata. See edit()"""
        with self.edit(rolename) as targets:
            if not isinstance(targets, Targets):
                raise AssertionError(f"Unexpected targets ({rolename}) type")
            yield targets

    def root(self) -> Root:
        """Read current root metadata"""
        root = self.open(Root.type).signed
        if not isinstance(root, Root):
            raise AssertionError("Unexpected root type")
        return root

    def timestamp(self) -> Timestamp:
        """Read current timestamp metadata"""
        timestamp = self.open(Timestamp.type).signed
        if not isinstance(timestamp, Timestamp):
            raise AssertionError("Unexpected timestamp type")
        return timestamp

    def snapshot(self) -> Snapshot:
        """Read current snapshot metadata"""
        snapshot = self.open(Snapshot.type).signed
        if not isinstance(snapshot, Snapshot):
            raise AssertionError("Unexpected snapshot type")
        return snapshot

    def targets(self, rolename: str = Targets.type) -> Targets:
        """Read current targets metadata"""
        targets = self.open(rolename).signed
        if not isinstance(targets, Targets):
            raise AssertionError("Unexpected targets type")
        return targets

    def do_snapshot(
        self, force: bool = False
    ) -> tuple[bool, dict[str, MetaFile]]:
        """Update snapshot meta information

        Updates the snapshot meta information according to current targets
        metadata state and the current snapshot meta information.

        Arguments:
            force: should new snapshot version be created even if meta
                information would not change?

        Returns: Tuple of
            - True if snapshot was created, False if not
            - MetaFiles for targets versions removed from snapshot meta
        """

        # Snapshot update is needed if
        # * any targets files are not yet in snapshot or
        # * any targets version is incorrect
        update_version = force
        removed: dict[str, MetaFile] = {}

        root = self.root()
        snapshot_md = self.open(Snapshot.type)

        try:
            root.verify_delegate(
                Snapshot.type,
                snapshot_md.signed_bytes,
                snapshot_md.signatures,
            )
        except UnsignedMetadataError:
            update_version = True

        with self.edit_snapshot() as snapshot:
            for keyname, new_meta in self.targets_infos.items():
                if keyname not in snapshot.meta:
                    update_version = True
                    snapshot.meta[keyname] = deepcopy(new_meta)
                    continue

                old_meta = snapshot.meta[keyname]
                if new_meta.version < old_meta.version:
                    raise ValueError(f"{keyname} version rollback")
                if new_meta.version > old_meta.version:
                    update_version = True
                    snapshot.meta[keyname] = deepcopy(new_meta)
                    removed[keyname] = old_meta

            if not update_version:
                # prevent edit_snapshot() from storing a new version
                raise AbortEdit("Skip snapshot: No targets version changes")

        if not update_version:
            # this is reachable as edit_snapshot() handles AbortEdit
            logger.debug("Snapshot update not needed")  # type: ignore[unreachable]
        else:
            logger.debug("Snapshot v%d", snapshot.version)

        return update_version, removed

    def do_timestamp(self, force: bool = False) -> tuple[bool, MetaFile | None]:
        """Update timestamp meta information

        Updates timestamp according to current snapshot state

        Returns: Tuple of
            - True if timestamp was created, False if not
            - MetaFile for snapshot version removed from timestamp (if any)
        """
        update_version = force
        removed = None

        root = self.root()
        timestamp_md = self.open(Timestamp.type)

        try:
            root.verify_delegate(
                Timestamp.type,
                timestamp_md.signed_bytes,
                timestamp_md.signatures,
            )
        except UnsignedMetadataError:
            update_version = True

        with self.edit_timestamp() as timestamp:
            if self.snapshot_info.version < timestamp.snapshot_meta.version:
                raise ValueError("snapshot version rollback")

            if self.snapshot_info.version > timestamp.snapshot_meta.version:
                update_version = True
                removed = timestamp.snapshot_meta
                timestamp.snapshot_meta = deepcopy(self.snapshot_info)

            if not update_version:
                raise AbortEdit("Skip timestamp: No snapshot version changes")

        if not update_version:
            # this is reachable as edit_timestamp() handles AbortEdit
            logger.debug("Timestamp update not needed")  # type: ignore[unreachable]
        else:
            logger.debug("Timestamp v%d", timestamp.version)
        return update_version, removed



# Copyright 2025, the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Test ngclient ProxyEnvironment"""

from __future__ import annotations

import sys
import unittest
from unittest.mock import Mock, patch

from urllib3 import PoolManager, ProxyManager

from tests import utils
from tuf.ngclient._internal.proxy import ProxyEnvironment


class TestProxyEnvironment(unittest.TestCase):
    """Test ngclient ProxyEnvironment implementation

    These tests use the ProxyEnvironment.get_pool_manager() endpoint and then
    look at the ProxyEnvironment._poolmanagers dict keys to decide if the result
    is correct.

    The test environment is changed via mocking getproxies(): this is a urllib
    method that returns a dict with the proxy environment variable contents.

    Testing ProxyEnvironment.request() would possibly be better but far more
    difficult: the current test implementation does not require actually setting up
    all of the different proxies.
    """

    def assert_pool_managers(
        self, env: ProxyEnvironment, expected: list[str | None]
    ) -> None:
        # Pool managers have the expected proxy urls
        self.assertEqual(list(env._pool_managers.keys()), expected)

        # Pool manager types are as expected
        for proxy_url, pool_manager in env._pool_managers.items():
            self.assertIsInstance(pool_manager, PoolManager)
            if proxy_url is not None:
                self.assertIsInstance(pool_manager, ProxyManager)

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_no_variables(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {}

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "subdomain.example.com")
        env.get_pool_manager("https", "differentsite.com")

        # There is a single pool manager (no proxies)
        self.assert_pool_managers(env, [None])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_proxy_set(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "https": "http://localhost:8888",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "differentsite.com")

        # There are two pool managers: A plain poolmanager and https proxymanager
        self.assert_pool_managers(env, [None, "http://localhost:8888"])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_proxies_set(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "http": "http://localhost:8888",
            "https": "http://localhost:9999",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "subdomain.example.com")
        env.get_pool_manager("https", "differentsite.com")

        # There are two pool managers: A http proxymanager and https proxymanager
        self.assert_pool_managers(
            env, ["http://localhost:8888", "http://localhost:9999"]
        )

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_no_proxy_set(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "http": "http://localhost:8888",
            "https": "http://localhost:9999",
            "no": "somesite.com, example.com, another.site.com",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")

        # There is a single pool manager (no proxies)
        self.assert_pool_managers(env, [None])

        env.get_pool_manager("http", "differentsite.com")
        env.get_pool_manager("https", "differentsite.com")

        # There are three pool managers: plain poolmanager for no_proxy domains,
        # http proxymanager and https proxymanager
        self.assert_pool_managers(
            env, [None, "http://localhost:8888", "http://localhost:9999"]
        )

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_no_proxy_subdomain_match(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "https": "http://localhost:9999",
            "no": "somesite.com, example.com, another.site.com",
        }

        env = ProxyEnvironment()

        # this should match example.com in no_proxy
        env.get_pool_manager("https", "subdomain.example.com")

        # There is a single pool manager (no proxies)
        self.assert_pool_managers(env, [None])

        # this should not match example.com in no_proxy
        env.get_pool_manager("https", "xexample.com")

        # There are two pool managers: plain poolmanager for no_proxy domains,
        # and a https proxymanager
        self.assert_pool_managers(env, [None, "http://localhost:9999"])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_no_proxy_wildcard(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "https": "http://localhost:8888",
            "no": "*",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "differentsite.com")
        env.get_pool_manager("https", "subdomain.example.com")

        # There is a single pool manager, no proxies
        self.assert_pool_managers(env, [None])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_no_proxy_leading_dot(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "https": "http://localhost:8888",
            "no": ".example.com",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "subdomain.example.com")

        # There is a single pool manager, no proxies
        self.assert_pool_managers(env, [None])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_all_proxy_set(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "all": "http://localhost:8888",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "subdomain.example.com")
        env.get_pool_manager("https", "differentsite.com")

        # There is a single proxy manager
        self.assert_pool_managers(env, ["http://localhost:8888"])

        # This urllib3 currently only handles http and https but let's test anyway
        env.get_pool_manager("file", None)

        # proxy manager and a plain pool manager
        self.assert_pool_managers(env, ["http://localhost:8888", None])

    @patch("tuf.ngclient._internal.proxy.getproxies")
    def test_all_proxy_and_no_proxy_set(self, mock_getproxies: Mock) -> None:
        mock_getproxies.return_value = {
            "all": "http://localhost:8888",
            "no": "somesite.com, example.com, another.site.com",
        }

        env = ProxyEnvironment()
        env.get_pool_manager("http", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "example.com")
        env.get_pool_manager("https", "subdomain.example.com")

        # There is a single pool manager (no proxies)
        self.assert_pool_managers(env, [None])

        env.get_pool_manager("http", "differentsite.com")
        env.get_pool_manager("https", "differentsite.com")

        # There are two pool managers: plain poolmanager for no_proxy domains and
        # one proxymanager
        self.assert_pool_managers(env, [None, "http://localhost:8888"])


if __name__ == "__main__":
    utils.configure_test_logging(sys.argv)
    unittest.main()



# Copyright 2021, New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Unit test for Urllib3Fetcher."""

import io
import logging
import math
import os
import sys
import tempfile
import unittest
from typing import ClassVar
from unittest.mock import Mock, patch

import urllib3

from tests import utils
from tuf.api import exceptions
from tuf.ngclient import Urllib3Fetcher

logger = logging.getLogger(__name__)


class TestFetcher(unittest.TestCase):
    """Test Urllib3Fetcher class."""

    server_process_handler: ClassVar[utils.TestServerProcess]

    @classmethod
    def setUpClass(cls) -> None:
        """
        Create a temporary file and launch a simple server in the
        current working directory.
        """
        cls.server_process_handler = utils.TestServerProcess(log=logger)

        cls.file_contents = b"junk data"
        cls.file_length = len(cls.file_contents)
        with tempfile.NamedTemporaryFile(
            dir=os.getcwd(), delete=False
        ) as cls.target_file:
            cls.target_file.write(cls.file_contents)

        cls.url_prefix = (
            f"http://{utils.TEST_HOST_ADDRESS}:"
            f"{cls.server_process_handler.port!s}"
        )
        target_filename = os.path.basename(cls.target_file.name)
        cls.url = f"{cls.url_prefix}/{target_filename}"

    @classmethod
    def tearDownClass(cls) -> None:
        # Stop server process and perform clean up.
        cls.server_process_handler.clean()
        os.remove(cls.target_file.name)

    def setUp(self) -> None:
        # Instantiate a concrete instance of FetcherInterface
        self.fetcher = Urllib3Fetcher()

    # Simple fetch.
    def test_fetch(self) -> None:
        with tempfile.TemporaryFile() as temp_file:
            for chunk in self.fetcher.fetch(self.url):
                temp_file.write(chunk)

            temp_file.seek(0)
            self.assertEqual(self.file_contents, temp_file.read())

    # URL data downloaded in more than one chunk
    def test_fetch_in_chunks(self) -> None:
        # Set a smaller chunk size to ensure that the file will be downloaded
        # in more than one chunk
        self.fetcher.chunk_size = 4

        # expected_chunks_count: 3 (depends on length of self.file_length)
        expected_chunks_count = math.ceil(
            self.file_length / self.fetcher.chunk_size
        )
        self.assertEqual(expected_chunks_count, 3)

        chunks_count = 0
        with tempfile.TemporaryFile() as temp_file:
            for chunk in self.fetcher.fetch(self.url):
                temp_file.write(chunk)
                chunks_count += 1

            temp_file.seek(0)
            self.assertEqual(self.file_contents, temp_file.read())
            # Check that we calculate chunks as expected
            self.assertEqual(chunks_count, expected_chunks_count)

    # Incorrect URL parsing
    def test_url_parsing(self) -> None:
        with self.assertRaises(exceptions.DownloadError):
            self.fetcher.fetch("http://invalid/")

    # File not found error
    def test_http_error(self) -> None:
        with self.assertRaises(exceptions.DownloadHTTPError) as cm:
            self.url = f"{self.url_prefix}/non-existing-path"
            self.fetcher.fetch(self.url)
        self.assertEqual(cm.exception.status_code, 404)

    # Response read timeout error
    @patch.object(urllib3.PoolManager, "request")
    def test_response_read_timeout(self, mock_session_get: Mock) -> None:
        mock_response = Mock()
        mock_response.status = 200
        attr = {
            "stream.side_effect": urllib3.exceptions.MaxRetryError(
                urllib3.connectionpool.ConnectionPool("localhost"),
                "",
                urllib3.exceptions.TimeoutError(),
            )
        }
        mock_response.configure_mock(**attr)
        mock_session_get.return_value = mock_response

        with self.assertRaises(exceptions.SlowRetrievalError):
            next(self.fetcher.fetch(self.url))
        mock_response.stream.assert_called_once()

    # Read/connect session timeout error
    @patch.object(
        urllib3.PoolManager,
        "request",
        side_effect=urllib3.exceptions.MaxRetryError(
            urllib3.connectionpool.ConnectionPool("localhost"),
            "",
            urllib3.exceptions.TimeoutError(),
        ),
    )
    def test_session_get_timeout(self, mock_session_get: Mock) -> None:
        with self.assertRaises(exceptions.SlowRetrievalError):
            self.fetcher.fetch(self.url)
        mock_session_get.assert_called_once()

    # Simple bytes download
    def test_download_bytes(self) -> None:
        data = self.fetcher.download_bytes(self.url, self.file_length)
        self.assertEqual(self.file_contents, data)

    # Download file smaller than required max_length
    def test_download_bytes_upper_length(self) -> None:
        data = self.fetcher.download_bytes(self.url, self.file_length + 4)
        self.assertEqual(self.file_contents, data)

    # Download a file bigger than expected
    def test_download_bytes_length_mismatch(self) -> None:
        with self.assertRaises(exceptions.DownloadLengthMismatchError):
            self.fetcher.download_bytes(self.url, self.file_length - 4)

    # Simple file download
    def test_download_file(self) -> None:
        with self.fetcher.download_file(
            self.url, self.file_length
        ) as temp_file:
            temp_file.seek(0, io.SEEK_END)
            self.assertEqual(self.file_length, temp_file.tell())

    # Download file smaller than required max_length
    def test_download_file_upper_length(self) -> None:
        with self.fetcher.download_file(
            self.url, self.file_length + 4
        ) as temp_file:
            temp_file.seek(0, io.SEEK_END)
            self.assertEqual(self.file_length, temp_file.tell())

    # Download a file bigger than expected
    def test_download_file_length_mismatch(self) -> None:
        with (
            self.assertRaises(exceptions.DownloadLengthMismatchError),
            self.fetcher.download_file(self.url, self.file_length - 4),
        ):
            pass  # we never get here as download_file() raises


# Run unit test.
if __name__ == "__main__":
    utils.configure_test_logging(sys.argv)
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join("..")))

import tuf

# -- Project information -----------------------------------------------------

project = "TUF"
copyright = "2021, New York University and the TUF contributors"
author = "New York University and the TUF contributors"


# -- General configuration ---------------------------------------------------

master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
]

autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_theme_options = {"logo_only": True}
html_logo = "tuf-horizontal-white.png"
html_favicon = "tuf-icon-32.png"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# -- Autodoc configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html

# Tone down the "tuf.api.metadata." repetition
add_module_names = False
python_use_unqualified_type_names = True

# Show typehints in argument doc lines, but not in signatures
autodoc_typehints = "description"

autodoc_default_options = {
    "members": True,
    "inherited-members": "Exception",  # excl. members inherited from 'Exception'
    "exclude-members": "to_dict, from_dict",
}

# Version
version = tuf.__version__



# Copyright 2021, New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Provides an implementation of ``FetcherInterface`` using the urllib3 HTTP
library.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

# Imports
import urllib3

import tuf
from tuf.api import exceptions
from tuf.ngclient._internal.proxy import ProxyEnvironment
from tuf.ngclient.fetcher import FetcherInterface

if TYPE_CHECKING:
    from collections.abc import Iterator

# Globals
logger = logging.getLogger(__name__)


# Classes
class Urllib3Fetcher(FetcherInterface):
    """An implementation of ``FetcherInterface`` based on the urllib3 library.

    Attributes:
        socket_timeout: Timeout in seconds, used for both initial connection
            delay and the maximum delay between bytes received.
        chunk_size: Chunk size in bytes used when downloading.
    """

    def __init__(
        self,
        socket_timeout: int = 30,
        chunk_size: int = 400000,
        app_user_agent: str | None = None,
    ) -> None:
        # Default settings
        self.socket_timeout: int = socket_timeout  # seconds
        self.chunk_size: int = chunk_size  # bytes

        # Create User-Agent.
        ua = f"python-tuf/{tuf.__version__}"
        if app_user_agent is not None:
            ua = f"{app_user_agent} {ua}"

        self._proxy_env = ProxyEnvironment(headers={"User-Agent": ua})

    def _fetch(self, url: str) -> Iterator[bytes]:
        """Fetch the contents of HTTP/HTTPS url from a remote server.

        Args:
            url: URL string that represents a file location.

        Raises:
            exceptions.SlowRetrievalError: Timeout occurs while receiving
                data.
            exceptions.DownloadHTTPError: HTTP error code is received.

        Returns:
            Bytes iterator
        """

        # Defer downloading the response body with preload_content=False.
        # Always set the timeout. This timeout value is interpreted by
        # urllib3 as:
        #  - connect timeout (max delay before first byte is received)
        #  - read (gap) timeout (max delay between bytes received)
        try:
            response = self._proxy_env.request(
                "GET",
                url,
                preload_content=False,
                timeout=urllib3.Timeout(self.socket_timeout),
            )
        except urllib3.exceptions.MaxRetryError as e:
            if isinstance(e.reason, urllib3.exceptions.TimeoutError):
                raise exceptions.SlowRetrievalError from e

        if response.status >= 400:
            response.close()
            raise exceptions.DownloadHTTPError(
                f"HTTP error occurred with status {response.status}",
                response.status,
            )

        return self._chunks(response)

    def _chunks(
        self, response: urllib3.response.BaseHTTPResponse
    ) -> Iterator[bytes]:
        """A generator function to be returned by fetch.

        This way the caller of fetch can differentiate between connection
        and actual data download.
        """

        try:
            yield from response.stream(self.chunk_size)
        except urllib3.exceptions.MaxRetryError as e:
            if isinstance(e.reason, urllib3.exceptions.TimeoutError):
                raise exceptions.SlowRetrievalError from e

        finally:
            response.release_conn()



# Copyright 2021, New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Test Updater class"""

from __future__ import annotations

import logging
import os
import shutil
import sys
import tempfile
import unittest
from collections.abc import Iterable
from typing import TYPE_CHECKING, ClassVar
from unittest.mock import MagicMock, patch

from securesystemslib.signer import Signer

from tests import utils
from tuf.api import exceptions
from tuf.api.metadata import (
    Metadata,
    Root,
    Snapshot,
    TargetFile,
    Targets,
    Timestamp,
)
from tuf.ngclient import Updater, UpdaterConfig

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

logger = logging.getLogger(__name__)


class TestUpdater(unittest.TestCase):
    """Test the Updater class from 'tuf/ngclient/updater.py'."""

    server_process_handler: ClassVar[utils.TestServerProcess]

    @classmethod
    def setUpClass(cls) -> None:
        cls.tmp_test_root_dir = tempfile.mkdtemp(dir=os.getcwd())

        # Launch a SimpleHTTPServer
        # Test cases will request metadata and target files that have been
        # pre-generated in 'tuf/tests/repository_data', and are copied to
        # CWD/tmp_test_root_dir/*
        cls.server_process_handler = utils.TestServerProcess(log=logger)

    @classmethod
    def tearDownClass(cls) -> None:
        # Cleans resources, flush the logged lines (if any) and remove test dir
        cls.server_process_handler.clean()
        shutil.rmtree(cls.tmp_test_root_dir)

    def setUp(self) -> None:
        # Create tmp test dir inside of tmp test root dir to independently serve
        # new repository files for each test. We delete all tmp dirs at once in
        # tearDownClass after the server has released all resources.
        self.tmp_test_dir = tempfile.mkdtemp(dir=self.tmp_test_root_dir)

        # Copy the original repository files provided in the test folder so that
        # any modifications are restricted to the copies.
        # The 'repository_data' directory is expected to exist in 'tuf.tests/'.
        original_repository_files = os.path.join(
            utils.TESTS_DIR, "repository_data"
        )

        original_repository = os.path.join(
            original_repository_files, "repository"
        )
        original_keystore = os.path.join(original_repository_files, "keystore")
        original_client = os.path.join(
            original_repository_files,
            "client",
            "test_repository1",
            "metadata",
            "current",
        )

        # Save references to the often-needed client repository directories.
        # Test cases need these references to access metadata and target files.
        self.repository_directory = os.path.join(
            self.tmp_test_dir, "repository"
        )
        self.keystore_directory = os.path.join(self.tmp_test_dir, "keystore")
        self.client_directory = os.path.join(self.tmp_test_dir, "client")

        # Copy the original 'repository', 'client', and 'keystore' directories
        # to the temporary repository the test cases can use.
        shutil.copytree(original_repository, self.repository_directory)
        shutil.copytree(original_client, self.client_directory)
        shutil.copytree(original_keystore, self.keystore_directory)

        # 'path/to/tmp/repository' -> 'localhost:8001/tmp/repository'.
        repository_basepath = self.repository_directory[len(os.getcwd()) :]
        url_prefix = (
            "http://"
            + utils.TEST_HOST_ADDRESS
            + ":"
            + str(self.server_process_handler.port)
            + repository_basepath.replace("\\", "/")
        )

        self.metadata_url = f"{url_prefix}/metadata/"
        self.targets_url = f"{url_prefix}/targets/"
        self.dl_dir = tempfile.mkdtemp(dir=self.tmp_test_dir)
        # Creating a repository instance.  The test cases will use this client
        # updater to refresh metadata, fetch target files, etc.
        self.updater = Updater(
            metadata_dir=self.client_directory,
            metadata_base_url=self.metadata_url,
            target_dir=self.dl_dir,
            target_base_url=self.targets_url,
            bootstrap=None,
        )

    def tearDown(self) -> None:
        # Logs stdout and stderr from the sever subprocess.
        self.server_process_handler.flush_log()

    def _modify_repository_root(
        self,
        modification_func: Callable[[Metadata], None],
        bump_version: bool = False,
    ) -> None:
        """Apply 'modification_func' to root and persist it."""
        role_path = os.path.join(
            self.repository_directory, "metadata", "root.json"
        )
        root = Metadata[Root].from_file(role_path)
        modification_func(root)
        if bump_version:
            root.signed.version += 1
        root_key_path = os.path.join(self.keystore_directory, "root_key")

        uri = f"file2:{root_key_path}"
        role = root.signed.get_delegated_role(Root.type)
        key = root.signed.get_key(role.keyids[0])
        signer = Signer.from_priv_key_uri(uri, key)

        root.sign(signer)
        root.to_file(
            os.path.join(self.repository_directory, "metadata", "root.json")
        )
        root.to_file(
            os.path.join(
                self.repository_directory,
                "metadata",
                f"{root.signed.version}.root.json",
            )
        )

    def _assert_files_exist(self, roles: Iterable[str]) -> None:
        """Assert that local metadata files match 'roles'"""
        expected_files = [f"{role}.json" for role in roles]
        found_files = [
            e.name for e in os.scandir(self.client_directory) if e.is_file()
        ]

        self.assertListEqual(sorted(found_files), sorted(expected_files))

    def test_refresh_and_download(self) -> None:
        # Test refresh without consistent targets - targets without hash prefix.

        # top-level targets are already in local cache (but remove others)
        os.remove(os.path.join(self.client_directory, "role1.json"))
        os.remove(os.path.join(self.client_directory, "role2.json"))

        # top-level metadata is in local directory already
        self.updater.refresh()
        self._assert_files_exist(
            [Root.type, Snapshot.type, Targets.type, Timestamp.type]
        )

        # Get targetinfos, assert that cache does not contain files
        info1 = self.updater.get_targetinfo("file1.txt")
        assert isinstance(info1, TargetFile)
        self._assert_files_exist(
            [Root.type, Snapshot.type, Targets.type, Timestamp.type]
        )

        # Get targetinfo for 'file3.txt' listed in the delegated role1
        info3 = self.updater.get_targetinfo("file3.txt")
        assert isinstance(info3, TargetFile)
        expected_files = [
            "role1",
            Root.type,
            Snapshot.type,
            Targets.type,
            Timestamp.type,
        ]
        self._assert_files_exist(expected_files)
        self.assertIsNone(self.updater.find_cached_target(info1))
        self.assertIsNone(self.updater.find_cached_target(info3))

        # Download files, assert that cache has correct files
        self.updater.download_target(info1)
        path = self.updater.find_cached_target(info1)
        self.assertEqual(path, os.path.join(self.dl_dir, info1.path))
        self.assertIsNone(self.updater.find_cached_target(info3))

        self.updater.download_target(info3)
        path = self.updater.find_cached_target(info1)
        self.assertEqual(path, os.path.join(self.dl_dir, info1.path))
        path = self.updater.find_cached_target(info3)
        self.assertEqual(path, os.path.join(self.dl_dir, info3.path))

    def test_refresh_with_only_local_root(self) -> None:
        os.remove(os.path.join(self.client_directory, "timestamp.json"))
        os.remove(os.path.join(self.client_directory, "snapshot.json"))
        os.remove(os.path.join(self.client_directory, "targets.json"))
        os.remove(os.path.join(self.client_directory, "role1.json"))
        os.remove(os.path.join(self.client_directory, "role2.json"))
        self._assert_files_exist([Root.type])

        self.updater.refresh()
        self._assert_files_exist(
            [Root.type, Snapshot.type, Targets.type, Timestamp.type]
        )

        # Get targetinfo for 'file3.txt' listed in the delegated role1
        self.updater.get_targetinfo("file3.txt")
        expected_files = [
            "role1",
            Root.type,
            Snapshot.type,
            Targets.type,
            Timestamp.type,
        ]
        self._assert_files_exist(expected_files)

    def test_implicit_refresh_with_only_local_root(self) -> None:
        os.remove(os.path.join(self.client_directory, "timestamp.json"))
        os.remove(os.path.join(self.client_directory, "snapshot.json"))
        os.remove(os.path.join(self.client_directory, "targets.json"))
        os.remove(os.path.join(self.client_directory, "role1.json"))
        os.remove(os.path.join(self.client_directory, "role2.json"))
        self._assert_files_exist(["root"])

        # Get targetinfo for 'file3.txt' listed in the delegated role1
        self.updater.get_targetinfo("file3.txt")
        expected_files = ["role1", "root", "snapshot", "targets", "timestamp"]
        self._assert_files_exist(expected_files)

    def test_both_target_urls_not_set(self) -> None:
        # target_base_url = None and Updater._target_base_url = None
        updater = Updater(
            self.client_directory,
            self.metadata_url,
            self.dl_dir,
            bootstrap=None,
        )
        info = TargetFile(1, {"sha256": ""}, "targetpath")
        with self.assertRaises(ValueError):
            updater.download_target(info)

    def test_no_target_dir_no_filepath(self) -> None:
        # filepath = None and Updater.target_dir = None
        updater = Updater(
            self.client_directory, self.metadata_url, bootstrap=None
        )
        info = TargetFile(1, {"sha256": ""}, "targetpath")
        with self.assertRaises(ValueError):
            updater.find_cached_target(info)
        with self.assertRaises(ValueError):
            updater.download_target(info)

    def test_external_targets_url(self) -> None:
        self.updater.refresh()
        info = self.updater.get_targetinfo("file1.txt")
        assert isinstance(info, TargetFile)

        self.updater.download_target(info, target_base_url=self.targets_url)

    def test_length_hash_mismatch(self) -> None:
        self.updater.refresh()
        targetinfo = self.updater.get_targetinfo("file1.txt")
        assert isinstance(targetinfo, TargetFile)

        length = targetinfo.length
        with self.assertRaises(exceptions.RepositoryError):
            targetinfo.length = 44
            self.updater.download_target(targetinfo)

        with self.assertRaises(exceptions.RepositoryError):
            targetinfo.length = length
            targetinfo.hashes = {"sha256": "abcd"}
            self.updater.download_target(targetinfo)

    def test_updating_root(self) -> None:
        # Bump root version, resign and refresh
        self._modify_repository_root(lambda _: None, bump_version=True)
        self.updater.refresh()
        self.assertEqual(self.updater._trusted_set.root.version, 2)

    def test_missing_targetinfo(self) -> None:
        self.updater.refresh()

        # Get targetinfo for non-existing file
        self.assertIsNone(self.updater.get_targetinfo("file33.txt"))

    @patch.object(os, "replace", wraps=os.replace)
    @patch.object(os, "remove", wraps=os.remove)
    def test_persist_metadata_fails(
        self, wrapped_remove: MagicMock, wrapped_replace: MagicMock
    ) -> None:
        # Testing that when write succeeds (the file is created) and replace
        # fails by throwing OSError, then the file will be deleted.
        wrapped_replace.side_effect = OSError()
        with self.assertRaises(OSError):
            self.updater._persist_metadata("target", b"data")

        wrapped_replace.assert_called_once()
        wrapped_remove.assert_called_once()

        # Assert that the created tempfile during writing is eventually deleted
        # or in other words, there is no temporary file left in the folder.
        for filename in os.listdir(self.updater._dir):
            self.assertFalse(filename.startswith("tmp"))

    def test_invalid_target_base_url(self) -> None:
        info = TargetFile(1, {"sha256": ""}, "targetpath")
        with self.assertRaises(exceptions.DownloadError):
            self.updater.download_target(
                info, target_base_url="http://invalid/"
            )

    def test_non_existing_target_file(self) -> None:
        info = TargetFile(1, {"sha256": ""}, "/non_existing_file.txt")
        # When non-existing target file is given, download fails with
        # "404 Client Error: File not found for url"
        with self.assertRaises(exceptions.DownloadHTTPError):
            self.updater.download_target(info)

    def test_user_agent(self) -> None:
        # test default
        self.updater.refresh()
        poolmgr = self.updater._fetcher._proxy_env.get_pool_manager(
            "http", "localhost"
        )
        ua = poolmgr.headers["User-Agent"]
        self.assertEqual(ua[:11], "python-tuf/")

        # test custom UA
        updater = Updater(
            self.client_directory,
            self.metadata_url,
            self.dl_dir,
            self.targets_url,
            config=UpdaterConfig(app_user_agent="MyApp/1.2.3"),
            bootstrap=None,
        )
        updater.refresh()
        poolmgr = updater._fetcher._proxy_env.get_pool_manager(
            "http", "localhost"
        )
        ua = poolmgr.headers["User-Agent"]

        self.assertEqual(ua[:23], "MyApp/1.2.3 python-tuf/")


if __name__ == "__main__":
    utils.configure_test_logging(sys.argv)
    unittest.main()



# Copyright 2021-2022 python-tuf contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Repository API: A helper library for repository implementations

This module is intended to make any "metadata editing" applications easier to
implement: this includes repository applications, CI integration components as
well as developer and signing tools.

The repository module is not considered part of the stable python-tuf API yet.
"""

from tuf.repository._repository import AbortEdit, Repository  # noqa: F401



#!/usr/bin/env python3

# Copyright 2012 - 2017, New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Simple HTTP server for python-tuf tests"""

import socketserver
from http.server import SimpleHTTPRequestHandler

# Allow reuse so you can re-run tests as often as you want even if the
# tests reuse ports. Otherwise TCP TIME-WAIT prevents reuse for ~1 minute
socketserver.TCPServer.allow_reuse_address = True

httpd = socketserver.TCPServer(("localhost", 0), SimpleHTTPRequestHandler)
port_message = "bind succeeded, server port is: " + str(httpd.server_address[1])
print(port_message)
httpd.serve_forever()



# Copyright 2020, TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""
<Program Name>
  test_utils.py

<Author>
  Martin Vrachev.

<Started>
  October 21, 2020.

<Copyright>
  See LICENSE-MIT OR LICENSE for licensing information.

<Purpose>
  Provide tests for some of the functions in utils.py module.
"""

import logging
import os
import socket
import sys
import unittest

from tests import utils

logger = logging.getLogger(__name__)


def can_connect(port: int) -> bool:
    """Check if a socket can connect on the given port"""
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect(("localhost", port))
        return True
    except Exception:  # noqa: BLE001
        return False
    finally:
        # The process will always enter in finally even after return.
        if sock:
            sock.close()


class TestServerProcess(unittest.TestCase):
    """Test functionality provided in TestServerProcess from tests/utils.py."""

    def test_simple_server_startup(self) -> None:
        # Test normal case
        server_process_handler = utils.TestServerProcess(log=logger)

        # Make sure we can connect to the server
        self.assertTrue(can_connect(server_process_handler.port))
        server_process_handler.clean()

    def test_cleanup(self) -> None:
        # Test normal case
        server_process_handler = utils.TestServerProcess(
            log=logger, server=os.path.join(utils.TESTS_DIR, "simple_server.py")
        )

        server_process_handler.clean()

        # Check if the process has successfully been killed.
        self.assertFalse(server_process_handler.is_process_running())

    def test_server_exit_before_timeout(self) -> None:
        with self.assertRaises(utils.TestServerProcessError):
            utils.TestServerProcess(logger, server="non_existing_server.py")

        # Test starting a server which immediately exits."
        with self.assertRaises(utils.TestServerProcessError):
            utils.TestServerProcess(logger, server="fast_server_exit.py")


if __name__ == "__main__":
    utils.configure_test_logging(sys.argv)
    unittest.main()



# Copyright 2021-2022 python-tuf contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""A Repository implementation for maintainer and developer tools"""

from __future__ import annotations

import contextlib
import copy
import json
import logging
import os
from datetime import datetime, timedelta, timezone

from securesystemslib.signer import CryptoSigner, Signer
from urllib3 import request

from tuf.api.exceptions import RepositoryError
from tuf.api.metadata import Metadata, MetaFile, TargetFile, Targets
from tuf.api.serialization.json import JSONSerializer
from tuf.ngclient import Updater
from tuf.repository import Repository

logger = logging.getLogger(__name__)


class LocalRepository(Repository):
    """A repository implementation that fetches data from a remote repository

    This implementation fetches metadata from a remote repository, potentially
    creates new versions of metadata, and submits to the remote repository.

    ngclient Updater is used to fetch metadata from remote server: this is good
    because we want to make sure the metadata we modify is verified, but also
    bad because we need some hacks to access the Updaters metadata.
    """

    expiry_period = timedelta(days=1)

    def __init__(self, metadata_dir: str, key_dir: str, base_url: str):
        self.key_dir = key_dir
        if not os.path.isdir(self.key_dir):
            os.makedirs(self.key_dir)

        self.base_url = base_url

        self.updater = Updater(
            metadata_dir=metadata_dir,
            metadata_base_url=f"{base_url}/metadata/",
            bootstrap=None,
        )
        self.updater.refresh()

    @property
    def targets_infos(self) -> dict[str, MetaFile]:
        raise NotImplementedError  # we never call snapshot

    @property
    def snapshot_info(self) -> MetaFile:
        raise NotImplementedError  # we never call timestamp

    def open(self, role: str) -> Metadata:
        """Return cached (or fetched) metadata"""

        # if there is a metadata version fetched from remote, use that
        # HACK: access Updater internals
        trusted_set = self.updater._trusted_set  # noqa: SLF001
        if role in trusted_set:
            # NOTE: The original signature wrapper (Metadata) was verified and
            # discarded upon inclusion in the trusted set. It is safe to use
            # a fresh wrapper. `close` will override existing signatures anyway.
            return Metadata(copy.deepcopy(trusted_set[role]))

        # otherwise we're creating metadata from scratch
        md = Metadata(Targets())
        # this makes version bumping in close() simpler
        md.signed.version = 0
        return md

    def close(self, role_name: str, md: Metadata) -> None:
        """Store a version of metadata. Handle version bumps, expiry, signing"""
        targets = self.targets()
        role = targets.get_delegated_role(role_name)
        public_key = targets.get_key(role.keyids[0])
        uri = f"file2:{self.key_dir}/{role_name}"

        signer = Signer.from_priv_key_uri(uri, public_key)

        md.signed.version += 1
        md.signed.expires = datetime.now(timezone.utc) + self.expiry_period

        md.sign(signer, append=False)

        # Upload using "api/role"
        uri = f"{self.base_url}/api/role/{role_name}"
        r = request("POST", uri, body=md.to_bytes(JSONSerializer()), timeout=5)
        if r.status != 200:
            raise RuntimeError(f"HTTP error {r.status}")

    def add_target(self, role: str, targetpath: str) -> bool:
        """Add target to roles metadata and submit new metadata version"""

        # HACK: make sure we have the roles metadata in updater._trusted_set
        # (or that we're publishing the first version)
        # HACK: Assume RepositoryError is because we're just publishing version
        # 1 (so the roles metadata does not exist on server yet)
        with contextlib.suppress(RepositoryError):
            self.updater.get_targetinfo(targetpath)

        data = bytes(targetpath, "utf-8")
        targetfile = TargetFile.from_data(targetpath, data)
        try:
            with self.edit_targets(role) as delegated:
                delegated.targets[targetpath] = targetfile

        except Exception as e:  # noqa: BLE001
            print(f"Failed to submit new {role} with added target: {e}")
            return False

        print(f"Uploaded role {role} v{delegated.version}")
        return True

    def add_delegation(self, role: str) -> bool:
        """Use the (unauthenticated) delegation adding API endpoint"""
        signer = CryptoSigner.generate_ecdsa()

        data = {signer.public_key.keyid: signer.public_key.to_dict()}
        url = f"{self.base_url}/api/delegation/{role}"
        r = request("POST", url, body=json.dumps(data), timeout=5)
        if r.status != 200:
            print(f"delegation failed with {r}")
            return False

        # Store the private key using rolename as filename
        with open(f"{self.key_dir}/{role}", "wb") as f:
            f.write(signer.private_bytes)

        print(f"Uploaded new delegation, stored key in {self.key_dir}/{role}")
        return True



# Copyright the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0


"""Helper classes for low-level Metadata API."""

from __future__ import annotations

import abc
import fnmatch
import hashlib
import io
import logging
import sys
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    TypeVar,
)

from securesystemslib import exceptions as sslib_exceptions
from securesystemslib.signer import Key, Signature

from tuf.api.exceptions import LengthOrHashMismatchError, UnsignedMetadataError

if TYPE_CHECKING:
    from collections.abc import Iterator

_ROOT = "root"
_SNAPSHOT = "snapshot"
_TARGETS = "targets"
_TIMESTAMP = "timestamp"

_DEFAULT_HASH_ALGORITHM = "sha256"
_BLAKE_HASH_ALGORITHM = "blake2b-256"

# We aim to support SPECIFICATION_VERSION and require the input metadata
# files to have the same major version (the first number) as ours.
SPECIFICATION_VERSION = ["1", "0", "31"]
TOP_LEVEL_ROLE_NAMES = {_ROOT, _TIMESTAMP, _SNAPSHOT, _TARGETS}

logger = logging.getLogger(__name__)

# T is a Generic type constraint for container payloads
T = TypeVar("T", "Root", "Timestamp", "Snapshot", "Targets")


def _get_digest(algo: str) -> Any:  # noqa: ANN401
    """New digest helper to support custom "blake2b-256" algo name."""
    if algo == _BLAKE_HASH_ALGORITHM:
        return hashlib.blake2b(digest_size=32)

    return hashlib.new(algo)


def _hash_bytes(data: bytes, algo: str) -> str:
    """Returns hexdigest for data using algo."""
    digest = _get_digest(algo)
    digest.update(data)

    return digest.hexdigest()


def _hash_file(f: IO[bytes], algo: str) -> str:
    """Returns hexdigest for file using algo."""
    f.seek(0)
    if sys.version_info >= (3, 11):
        digest = hashlib.file_digest(f, lambda: _get_digest(algo))  # type: ignore[arg-type]

    else:
        # Fallback for older Pythons. Chunk size is taken from the previously
        # used and now deprecated `securesystemslib.hash.digest_fileobject`.
        digest = _get_digest(algo)
        for chunk in iter(lambda: f.read(4096), b""):
            digest.update(chunk)

    return digest.hexdigest()


class Signed(metaclass=abc.ABCMeta):
    """A base class for the signed part of TUF metadata.

    Objects with base class Signed are usually included in a ``Metadata`` object
    on the signed attribute. This class provides attributes and methods that
    are common for all TUF metadata types (roles).

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        version: Metadata version number. If None, then 1 is assigned.
        spec_version: Supported TUF specification version. If None, then the
            version currently supported by the library is assigned.
        expires: Metadata expiry date in UTC timezone. If None, then current
            date and time is assigned.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError: Invalid arguments.
    """

    # type is required for static reference without changing the API
    type: ClassVar[str] = "signed"

    # _type and type are identical: 1st replicates file format, 2nd passes lint
    @property
    def _type(self) -> str:
        return self.type

    @property
    def expires(self) -> datetime:
        """Get the metadata expiry date."""
        return self._expires

    @expires.setter
    def expires(self, value: datetime) -> None:
        """Set the metadata expiry date.

        # Use 'datetime' module to e.g. expire in seven days from now
        obj.expires = now(timezone.utc) + timedelta(days=7)
        """
        self._expires = value.replace(microsecond=0)
        if self._expires.tzinfo is None:
            # Naive datetime: just make it UTC
            self._expires = self._expires.replace(tzinfo=timezone.utc)
        elif self._expires.tzinfo != timezone.utc:
            raise ValueError(f"Expected tz UTC, not {self._expires.tzinfo}")

    # NOTE: Signed is a stupid name, because this might not be signed yet, but
    # we keep it to match spec terminology (I often refer to this as "payload",
    # or "inner metadata")
    def __init__(
        self,
        version: int | None,
        spec_version: str | None,
        expires: datetime | None,
        unrecognized_fields: dict[str, Any] | None,
    ):
        if spec_version is None:
            spec_version = ".".join(SPECIFICATION_VERSION)
        # Accept semver (X.Y.Z) but also X.Y for legacy compatibility
        spec_list = spec_version.split(".")
        if len(spec_list) not in [2, 3] or not all(
            el.isdigit() for el in spec_list
        ):
            raise ValueError(f"Failed to parse spec_version {spec_version}")

        # major version must match
        if spec_list[0] != SPECIFICATION_VERSION[0]:
            raise ValueError(f"Unsupported spec_version {spec_version}")

        self.spec_version = spec_version

        self.expires = expires or datetime.now(timezone.utc)

        if version is None:
            version = 1
        elif version <= 0:
            raise ValueError(f"version must be > 0, got {version}")
        self.version = version

        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Signed):
            return False

        return (
            self.type == other.type
            and self.version == other.version
            and self.spec_version == other.spec_version
            and self.expires == other.expires
            and self.unrecognized_fields == other.unrecognized_fields
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.type,
                self.version,
                self.spec_version,
                self.expires,
                self.unrecognized_fields,
            )
        )

    @abc.abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Serialize and return a dict representation of self."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, signed_dict: dict[str, Any]) -> Signed:
        """Deserialization helper, creates object from json/dict
        representation.
        """
        raise NotImplementedError

    @classmethod
    def _common_fields_from_dict(
        cls, signed_dict: dict[str, Any]
    ) -> tuple[int, str, datetime]:
        """Return common fields of ``Signed`` instances from the passed dict
        representation, and returns an ordered list to be passed as leading
        positional arguments to a subclass constructor.

        See ``{Root, Timestamp, Snapshot, Targets}.from_dict``
        methods for usage.

        """
        _type = signed_dict.pop("_type")
        if _type != cls.type:
            raise ValueError(f"Expected type {cls.type}, got {_type}")

        version = signed_dict.pop("version")
        spec_version = signed_dict.pop("spec_version")
        expires_str = signed_dict.pop("expires")
        # Convert 'expires' TUF metadata string to a datetime object, which is
        # what the constructor expects and what we store. The inverse operation
        # is implemented in '_common_fields_to_dict'.
        expires = datetime.strptime(expires_str, "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=timezone.utc
        )

        return version, spec_version, expires

    def _common_fields_to_dict(self) -> dict[str, Any]:
        """Return a dict representation of common fields of
        ``Signed`` instances.

        See ``{Root, Timestamp, Snapshot, Targets}.to_dict`` methods for usage.

        """
        return {
            "_type": self._type,
            "version": self.version,
            "spec_version": self.spec_version,
            "expires": self.expires.strftime("%Y-%m-%dT%H:%M:%SZ"),
            **self.unrecognized_fields,
        }

    def is_expired(self, reference_time: datetime | None = None) -> bool:
        """Check metadata expiration against a reference time.

        Args:
            reference_time: Time to check expiration date against. A naive
                datetime in UTC expected. Default is current UTC date and time.

        Returns:
            ``True`` if expiration time is less than the reference time.
        """
        if reference_time is None:
            reference_time = datetime.now(timezone.utc)

        return reference_time >= self.expires


class Role:
    """Container that defines which keys are required to sign roles metadata.

    Role defines how many keys are required to successfully sign the roles
    metadata, and which keys are accepted.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        keyids: Roles signing key identifiers.
        threshold: Number of keys required to sign this role's metadata.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError: Invalid arguments.
    """

    def __init__(
        self,
        keyids: list[str],
        threshold: int,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        if len(set(keyids)) != len(keyids):
            raise ValueError(f"Nonunique keyids: {keyids}")
        if threshold < 1:
            raise ValueError("threshold should be at least 1!")
        self.keyids = keyids
        self.threshold = threshold
        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Role):
            return False

        return (
            self.keyids == other.keyids
            and self.threshold == other.threshold
            and self.unrecognized_fields == other.unrecognized_fields
        )

    def __hash__(self) -> int:
        return hash((self.keyids, self.threshold, self.unrecognized_fields))

    @classmethod
    def from_dict(cls, role_dict: dict[str, Any]) -> Role:
        """Create ``Role`` object from its json/dict representation.

        Raises:
            ValueError, KeyError: Invalid arguments.
        """
        keyids = role_dict.pop("keyids")
        threshold = role_dict.pop("threshold")
        # All fields left in the role_dict are unrecognized.
        return cls(keyids, threshold, role_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of self."""
        return {
            "keyids": self.keyids,
            "threshold": self.threshold,
            **self.unrecognized_fields,
        }


@dataclass
class VerificationResult:
    """Signature verification result for delegated role metadata.

    Attributes:
        threshold: Number of required signatures.
        signed: dict of keyid to Key, containing keys that have signed.
        unsigned: dict of keyid to Key, containing keys that have not signed.
    """

    threshold: int
    signed: dict[str, Key]
    unsigned: dict[str, Key]

    def __bool__(self) -> bool:
        return self.verified

    @property
    def verified(self) -> bool:
        """True if threshold of signatures is met."""
        return len(self.signed) >= self.threshold

    @property
    def missing(self) -> int:
        """Number of additional signatures required to reach threshold."""
        return max(0, self.threshold - len(self.signed))


@dataclass
class RootVerificationResult:
    """Signature verification result for root metadata.

    Root must be verified by itself and the previous root version. This
    dataclass represents both results. For the edge case of first version
    of root, these underlying results are identical.

    Note that `signed` and `unsigned` correctness requires the underlying
    VerificationResult keys to not conflict (no reusing the same keyid for
    different keys).

    Attributes:
        first: First underlying VerificationResult
        second: Second underlying VerificationResult
    """

    first: VerificationResult
    second: VerificationResult

    def __bool__(self) -> bool:
        return self.verified

    @property
    def verified(self) -> bool:
        """True if threshold of signatures is met in both underlying
        VerificationResults.
        """
        return self.first.verified and self.second.verified

    @property
    def signed(self) -> dict[str, Key]:
        """Dictionary of all signing keys that have signed, from both
        VerificationResults.
        """
        return self.first.signed | self.second.signed

    @property
    def unsigned(self) -> dict[str, Key]:
        """Dictionary of all signing keys that have not signed, from both
        VerificationResults.
        """
        return self.first.unsigned | self.second.unsigned


class _DelegatorMixin(metaclass=abc.ABCMeta):
    """Class that implements verify_delegate() for Root and Targets"""

    @abc.abstractmethod
    def get_delegated_role(self, delegated_role: str) -> Role:
        """Return the role object for the given delegated role.

        Raises ValueError if delegated_role is not actually delegated.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_key(self, keyid: str) -> Key:
        """Return the key object for the given keyid.

        Raises ValueError if key is not found.
        """
        raise NotImplementedError

    def get_verification_result(
        self,
        delegated_role: str,
        payload: bytes,
        signatures: dict[str, Signature],
    ) -> VerificationResult:
        """Return signature threshold verification result for delegated role.

        NOTE: Unlike `verify_delegate()` this method does not raise, if the
        role metadata is not fully verified.

        Args:
            delegated_role: Name of the delegated role to verify
            payload: Signed payload bytes for the delegated role
            signatures: Signatures over payload bytes

        Raises:
            ValueError: no delegation was found for ``delegated_role``.
        """
        role = self.get_delegated_role(delegated_role)

        signed = {}
        unsigned = {}

        for keyid in role.keyids:
            try:
                key = self.get_key(keyid)
            except ValueError:
                logger.info("No key for keyid %s", keyid)
                continue

            if keyid not in signatures:
                unsigned[keyid] = key
                logger.info("No signature for keyid %s", keyid)
                continue

            sig = signatures[keyid]
            try:
                key.verify_signature(sig, payload)
                signed[keyid] = key
            except sslib_exceptions.UnverifiedSignatureError:
                unsigned[keyid] = key
                logger.info("Key %s failed to verify %s", keyid, delegated_role)

        return VerificationResult(role.threshold, signed, unsigned)

    def verify_delegate(
        self,
        delegated_role: str,
        payload: bytes,
        signatures: dict[str, Signature],
    ) -> None:
        """Verify signature threshold for delegated role.

        Verify that there are enough valid ``signatures`` over ``payload``, to
        meet the threshold of keys for ``delegated_role``, as defined by the
        delegator (``self``).

        Args:
            delegated_role: Name of the delegated role to verify
            payload: Signed payload bytes for the delegated role
            signatures: Signatures over payload bytes

        Raises:
            UnsignedMetadataError: ``delegated_role`` was not signed with
                required threshold of keys for ``role_name``.
            ValueError: no delegation was found for ``delegated_role``.
        """
        result = self.get_verification_result(
            delegated_role, payload, signatures
        )
        if not result:
            raise UnsignedMetadataError(
                f"{delegated_role} was signed by {len(result.signed)}/"
                f"{result.threshold} keys"
            )


class Root(Signed, _DelegatorMixin):
    """A container for the signed part of root metadata.

    Parameters listed below are also instance attributes.

    Args:
        version: Metadata version number. Default is 1.
        spec_version: Supported TUF specification version. Default is the
            version currently supported by the library.
        expires: Metadata expiry date. Default is current date and time.
        keys: Dictionary of keyids to Keys. Defines the keys used in ``roles``.
            Default is empty dictionary.
        roles: Dictionary of role names to Roles. Defines which keys are
            required to sign the metadata for a specific role. Default is
            a dictionary of top level roles without keys and threshold of 1.
        consistent_snapshot: ``True`` if repository supports consistent
        snapshots. Default is True.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError: Invalid arguments.
    """

    type = _ROOT

    def __init__(
        self,
        version: int | None = None,
        spec_version: str | None = None,
        expires: datetime | None = None,
        keys: dict[str, Key] | None = None,
        roles: dict[str, Role] | None = None,
        consistent_snapshot: bool | None = True,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        super().__init__(version, spec_version, expires, unrecognized_fields)
        self.consistent_snapshot = consistent_snapshot
        self.keys = keys if keys is not None else {}

        if roles is None:
            roles = {r: Role([], 1) for r in TOP_LEVEL_ROLE_NAMES}
        elif set(roles) != TOP_LEVEL_ROLE_NAMES:
            raise ValueError("Role names must be the top-level metadata roles")
        self.roles = roles

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Root):
            return False

        return (
            super().__eq__(other)
            and self.keys == other.keys
            and self.roles == other.roles
            and self.consistent_snapshot == other.consistent_snapshot
        )

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self.keys,
                self.roles,
                self.consistent_snapshot,
                self.unrecognized_fields,
            )
        )

    @classmethod
    def from_dict(cls, signed_dict: dict[str, Any]) -> Root:
        """Create ``Root`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, TypeError: Invalid arguments.
        """
        common_args = cls._common_fields_from_dict(signed_dict)
        consistent_snapshot = signed_dict.pop("consistent_snapshot", None)
        keys = signed_dict.pop("keys")
        roles = signed_dict.pop("roles")

        for keyid, key_dict in keys.items():
            keys[keyid] = Key.from_dict(keyid, key_dict)
        for role_name, role_dict in roles.items():
            roles[role_name] = Role.from_dict(role_dict)

        # All fields left in the signed_dict are unrecognized.
        return cls(*common_args, keys, roles, consistent_snapshot, signed_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        root_dict = self._common_fields_to_dict()
        keys = {keyid: key.to_dict() for (keyid, key) in self.keys.items()}
        roles = {}
        for role_name, role in self.roles.items():
            roles[role_name] = role.to_dict()
        if self.consistent_snapshot is not None:
            root_dict["consistent_snapshot"] = self.consistent_snapshot

        root_dict.update(
            {
                "keys": keys,
                "roles": roles,
            }
        )
        return root_dict

    def add_key(self, key: Key, role: str) -> None:
        """Add new signing key for delegated role ``role``.

        Args:
            key: Signing key to be added for ``role``.
            role: Name of the role, for which ``key`` is added.

        Raises:
            ValueError: If the argument order is wrong or if ``role`` doesn't
                exist.
        """
        # Verify that our users are not using the old argument order.
        if isinstance(role, Key):
            raise ValueError("Role must be a string, not a Key instance")

        if role not in self.roles:
            raise ValueError(f"Role {role} doesn't exist")
        if key.keyid not in self.roles[role].keyids:
            self.roles[role].keyids.append(key.keyid)
        self.keys[key.keyid] = key

    def revoke_key(self, keyid: str, role: str) -> None:
        """Revoke key from ``role`` and updates the key store.

        Args:
            keyid: Identifier of the key to be removed for ``role``.
            role: Name of the role, for which a signing key is removed.

        Raises:
            ValueError: If ``role`` doesn't exist or if ``role`` doesn't include
                the key.
        """
        if role not in self.roles:
            raise ValueError(f"Role {role} doesn't exist")
        if keyid not in self.roles[role].keyids:
            raise ValueError(f"Key with id {keyid} is not used by {role}")
        self.roles[role].keyids.remove(keyid)
        for keyinfo in self.roles.values():
            if keyid in keyinfo.keyids:
                return

        del self.keys[keyid]

    def get_delegated_role(self, delegated_role: str) -> Role:
        """Return the role object for the given delegated role.

        Raises ValueError if delegated_role is not actually delegated.
        """
        if delegated_role not in self.roles:
            raise ValueError(f"Delegated role {delegated_role} not found")

        return self.roles[delegated_role]

    def get_key(self, keyid: str) -> Key:
        if keyid not in self.keys:
            raise ValueError(f"Key {keyid} not found")

        return self.keys[keyid]

    def get_root_verification_result(
        self,
        previous: Root | None,
        payload: bytes,
        signatures: dict[str, Signature],
    ) -> RootVerificationResult:
        """Return signature threshold verification result for two root roles.

        Verify root metadata with two roles (`self` and optionally `previous`).

        If the repository has no root role versions yet, `previous` can be left
        None. In all other cases, `previous` must be the previous version of
        the Root.

        NOTE: Unlike `verify_delegate()` this method does not raise, if the
        root metadata is not fully verified.

        Args:
            previous: The previous `Root` to verify payload with, or None
            payload: Signed payload bytes for root
            signatures: Signatures over payload bytes

        Raises:
            ValueError: no delegation was found for ``root`` or given Root
                versions are not sequential.
        """

        if previous is None:
            previous = self
        elif self.version != previous.version + 1:
            versions = f"v{previous.version} and v{self.version}"
            raise ValueError(
                f"Expected sequential root versions, got {versions}."
            )

        return RootVerificationResult(
            previous.get_verification_result(Root.type, payload, signatures),
            self.get_verification_result(Root.type, payload, signatures),
        )


class BaseFile:
    """A base class of ``MetaFile`` and ``TargetFile``.

    Encapsulates common static methods for length and hash verification.
    """

    @staticmethod
    def _verify_hashes(
        data: bytes | IO[bytes], expected_hashes: dict[str, str]
    ) -> None:
        """Verify that the hash of ``data`` matches ``expected_hashes``."""
        for algo, exp_hash in expected_hashes.items():
            try:
                if isinstance(data, bytes):
                    observed_hash = _hash_bytes(data, algo)
                else:
                    # if data is not bytes, assume it is a file object
                    observed_hash = _hash_file(data, algo)
            except (ValueError, TypeError) as e:
                raise LengthOrHashMismatchError(
                    f"Unsupported algorithm '{algo}'"
                ) from e

            if observed_hash != exp_hash:
                raise LengthOrHashMismatchError(
                    f"Observed hash {observed_hash} does not match "
                    f"expected hash {exp_hash}"
                )

    @staticmethod
    def _verify_length(data: bytes | IO[bytes], expected_length: int) -> None:
        """Verify that the length of ``data`` matches ``expected_length``."""
        if isinstance(data, bytes):
            observed_length = len(data)
        else:
            # if data is not bytes, assume it is a file object
            data.seek(0, io.SEEK_END)
            observed_length = data.tell()

        if observed_length != expected_length:
            raise LengthOrHashMismatchError(
                f"Observed length {observed_length} does not match "
                f"expected length {expected_length}"
            )

    @staticmethod
    def _validate_hashes(hashes: dict[str, str]) -> None:
        if not hashes:
            raise ValueError("Hashes must be a non empty dictionary")
        for key, value in hashes.items():
            if not (isinstance(key, str) and isinstance(value, str)):
                raise TypeError("Hashes items must be strings")

    @staticmethod
    def _validate_length(length: int) -> None:
        if length < 0:
            raise ValueError(f"Length must be >= 0, got {length}")

    @staticmethod
    def _get_length_and_hashes(
        data: bytes | IO[bytes], hash_algorithms: list[str] | None
    ) -> tuple[int, dict[str, str]]:
        """Calculate length and hashes of ``data``."""
        if isinstance(data, bytes):
            length = len(data)
        else:
            data.seek(0, io.SEEK_END)
            length = data.tell()

        hashes = {}

        if hash_algorithms is None:
            hash_algorithms = [_DEFAULT_HASH_ALGORITHM]

        for algorithm in hash_algorithms:
            try:
                if isinstance(data, bytes):
                    hashes[algorithm] = _hash_bytes(data, algorithm)
                else:
                    hashes[algorithm] = _hash_file(data, algorithm)
            except (ValueError, TypeError) as e:
                raise ValueError(f"Unsupported algorithm '{algorithm}'") from e

        return (length, hashes)


class MetaFile(BaseFile):
    """A container with information about a particular metadata file.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        version: Version of the metadata file.
        length: Length of the metadata file in bytes.
        hashes: Dictionary of hash algorithm names to hashes of the metadata
            file content.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError, TypeError: Invalid arguments.
    """

    def __init__(
        self,
        version: int = 1,
        length: int | None = None,
        hashes: dict[str, str] | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        if version <= 0:
            raise ValueError(f"Metafile version must be > 0, got {version}")
        if length is not None:
            self._validate_length(length)
        if hashes is not None:
            self._validate_hashes(hashes)

        self.version = version
        self.length = length
        self.hashes = hashes
        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MetaFile):
            return False

        return (
            self.version == other.version
            and self.length == other.length
            and self.hashes == other.hashes
            and self.unrecognized_fields == other.unrecognized_fields
        )

    def __hash__(self) -> int:
        return hash(
            (self.version, self.length, self.hashes, self.unrecognized_fields)
        )

    @classmethod
    def from_dict(cls, meta_dict: dict[str, Any]) -> MetaFile:
        """Create ``MetaFile`` object from its json/dict representation.

        Raises:
            ValueError, KeyError: Invalid arguments.
        """
        version = meta_dict.pop("version")
        length = meta_dict.pop("length", None)
        hashes = meta_dict.pop("hashes", None)

        # All fields left in the meta_dict are unrecognized.
        return cls(version, length, hashes, meta_dict)

    @classmethod
    def from_data(
        cls,
        version: int,
        data: bytes | IO[bytes],
        hash_algorithms: list[str],
    ) -> MetaFile:
        """Creates MetaFile object from bytes.
        This constructor should only be used if hashes are wanted.
        By default, MetaFile(ver) should be used.
        Args:
            version: Version of the metadata file.
            data: Metadata bytes that the metafile represents.
            hash_algorithms: Hash algorithms to create the hashes with. If not
            specified, "sha256" is used.

        Raises:
            ValueError: The hash algorithms list contains an unsupported
            algorithm.
        """
        length, hashes = cls._get_length_and_hashes(data, hash_algorithms)
        return cls(version, length, hashes)

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of self."""
        res_dict: dict[str, Any] = {
            "version": self.version,
            **self.unrecognized_fields,
        }

        if self.length is not None:
            res_dict["length"] = self.length

        if self.hashes is not None:
            res_dict["hashes"] = self.hashes

        return res_dict

    def verify_length_and_hashes(self, data: bytes | IO[bytes]) -> None:
        """Verify that the length and hashes of ``data`` match expected values.

        Args:
            data: File object or its content in bytes.

        Raises:
            LengthOrHashMismatchError: Calculated length or hashes do not
                match expected values or hash algorithm is not supported.
        """
        if self.length is not None:
            self._verify_length(data, self.length)

        if self.hashes is not None:
            self._verify_hashes(data, self.hashes)


class Timestamp(Signed):
    """A container for the signed part of timestamp metadata.

    TUF file format uses a dictionary to contain the snapshot information:
    this is not the case with ``Timestamp.snapshot_meta`` which is a
    ``MetaFile``.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        version: Metadata version number. Default is 1.
        spec_version: Supported TUF specification version. Default is the
            version currently supported by the library.
        expires: Metadata expiry date. Default is current date and time.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API
        snapshot_meta: Meta information for snapshot metadata. Default is a
            MetaFile with version 1.

    Raises:
        ValueError: Invalid arguments.
    """

    type = _TIMESTAMP

    def __init__(
        self,
        version: int | None = None,
        spec_version: str | None = None,
        expires: datetime | None = None,
        snapshot_meta: MetaFile | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        super().__init__(version, spec_version, expires, unrecognized_fields)
        self.snapshot_meta = snapshot_meta or MetaFile(1)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return False

        return (
            super().__eq__(other) and self.snapshot_meta == other.snapshot_meta
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.snapshot_meta))

    @classmethod
    def from_dict(cls, signed_dict: dict[str, Any]) -> Timestamp:
        """Create ``Timestamp`` object from its json/dict representation.

        Raises:
            ValueError, KeyError: Invalid arguments.
        """
        common_args = cls._common_fields_from_dict(signed_dict)
        meta_dict = signed_dict.pop("meta")
        snapshot_meta = MetaFile.from_dict(meta_dict["snapshot.json"])
        # All fields left in the timestamp_dict are unrecognized.
        return cls(*common_args, snapshot_meta, signed_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        res_dict = self._common_fields_to_dict()
        res_dict["meta"] = {"snapshot.json": self.snapshot_meta.to_dict()}
        return res_dict


class Snapshot(Signed):
    """A container for the signed part of snapshot metadata.

    Snapshot contains information about all target Metadata files.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        version: Metadata version number. Default is 1.
        spec_version: Supported TUF specification version. Default is the
            version currently supported by the library.
        expires: Metadata expiry date. Default is current date and time.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API
        meta: Dictionary of targets filenames to ``MetaFile`` objects. Default
            is a dictionary with a Metafile for "snapshot.json" version 1.

    Raises:
        ValueError: Invalid arguments.
    """

    type = _SNAPSHOT

    def __init__(
        self,
        version: int | None = None,
        spec_version: str | None = None,
        expires: datetime | None = None,
        meta: dict[str, MetaFile] | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        super().__init__(version, spec_version, expires, unrecognized_fields)
        self.meta = meta if meta is not None else {"targets.json": MetaFile(1)}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Snapshot):
            return False

        return super().__eq__(other) and self.meta == other.meta

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.meta))

    @classmethod
    def from_dict(cls, signed_dict: dict[str, Any]) -> Snapshot:
        """Create ``Snapshot`` object from its json/dict representation.

        Raises:
            ValueError, KeyError: Invalid arguments.
        """
        common_args = cls._common_fields_from_dict(signed_dict)
        meta_dicts = signed_dict.pop("meta")
        meta = {}
        for meta_path, meta_dict in meta_dicts.items():
            meta[meta_path] = MetaFile.from_dict(meta_dict)
        # All fields left in the snapshot_dict are unrecognized.
        return cls(*common_args, meta, signed_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        snapshot_dict = self._common_fields_to_dict()
        meta_dict = {}
        for meta_path, meta_info in self.meta.items():
            meta_dict[meta_path] = meta_info.to_dict()

        snapshot_dict["meta"] = meta_dict
        return snapshot_dict


class DelegatedRole(Role):
    """A container with information about a delegated role.

    A delegation can happen in two ways:

        - ``paths`` is set: delegates targets matching any path pattern in
          ``paths``
        - ``path_hash_prefixes`` is set: delegates targets whose target path
          hash starts with any of the prefixes in ``path_hash_prefixes``

        ``paths`` and ``path_hash_prefixes`` are mutually exclusive:
        both cannot be set, at least one of them must be set.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        name: Delegated role name.
        keyids: Delegated role signing key identifiers.
        threshold: Number of keys required to sign this role's metadata.
        terminating: ``True`` if this delegation terminates a target lookup.
        paths: Path patterns. See note above.
        path_hash_prefixes: Hash prefixes. See note above.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API.

    Raises:
        ValueError: Invalid arguments.
    """

    def __init__(
        self,
        name: str,
        keyids: list[str],
        threshold: int,
        terminating: bool,
        paths: list[str] | None = None,
        path_hash_prefixes: list[str] | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        super().__init__(keyids, threshold, unrecognized_fields)
        self.name = name
        self.terminating = terminating
        exclusive_vars = [paths, path_hash_prefixes]
        if sum(1 for var in exclusive_vars if var is not None) != 1:
            raise ValueError(
                "Only one of (paths, path_hash_prefixes) must be set"
            )

        if paths is not None and any(not isinstance(p, str) for p in paths):
            raise ValueError("Paths must be strings")
        if path_hash_prefixes is not None and any(
            not isinstance(p, str) for p in path_hash_prefixes
        ):
            raise ValueError("Path_hash_prefixes must be strings")

        self.paths = paths
        self.path_hash_prefixes = path_hash_prefixes

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DelegatedRole):
            return False

        return (
            super().__eq__(other)
            and self.name == other.name
            and self.terminating == other.terminating
            and self.paths == other.paths
            and self.path_hash_prefixes == other.path_hash_prefixes
        )

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self.name,
                self.terminating,
                self.path,
                self.path_hash_prefixes,
            )
        )

    @classmethod
    def from_dict(cls, role_dict: dict[str, Any]) -> DelegatedRole:
        """Create ``DelegatedRole`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, TypeError: Invalid arguments.
        """
        name = role_dict.pop("name")
        keyids = role_dict.pop("keyids")
        threshold = role_dict.pop("threshold")
        terminating = role_dict.pop("terminating")
        paths = role_dict.pop("paths", None)
        path_hash_prefixes = role_dict.pop("path_hash_prefixes", None)
        # All fields left in the role_dict are unrecognized.
        return cls(
            name,
            keyids,
            threshold,
            terminating,
            paths,
            path_hash_prefixes,
            role_dict,
        )

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        base_role_dict = super().to_dict()
        res_dict = {
            "name": self.name,
            "terminating": self.terminating,
            **base_role_dict,
        }
        if self.paths is not None:
            res_dict["paths"] = self.paths
        elif self.path_hash_prefixes is not None:
            res_dict["path_hash_prefixes"] = self.path_hash_prefixes
        return res_dict

    @staticmethod
    def _is_target_in_pathpattern(targetpath: str, pathpattern: str) -> bool:
        """Determine whether ``targetpath`` matches the ``pathpattern``."""
        # We need to make sure that targetpath and pathpattern are pointing to
        # the same directory as fnmatch doesn't threat "/" as a special symbol.
        target_parts = targetpath.split("/")
        pattern_parts = pathpattern.split("/")
        if len(target_parts) != len(pattern_parts):
            return False

        # Every part in the pathpattern could include a glob pattern, that's why
        # each of the target and pathpattern parts should match.
        for target, pattern in zip(target_parts, pattern_parts, strict=True):
            if not fnmatch.fnmatchcase(target, pattern):
                return False

        return True

    def is_delegated_path(self, target_filepath: str) -> bool:
        """Determine whether the given ``target_filepath`` is in one of
        the paths that ``DelegatedRole`` is trusted to provide.

        The ``target_filepath`` and the ``DelegatedRole`` paths are expected to
        be in their canonical forms, so e.g. "a/b" instead of "a//b" . Only "/"
        is supported as target path separator. Leading separators are not
        handled as special cases (see `TUF specification on targetpath
        <https://theupdateframework.github.io/specification/latest/#targetpath>`_).

        Args:
            target_filepath: URL path to a target file, relative to a base
                targets URL.
        """

        if self.path_hash_prefixes is not None:
            # Calculate the hash of the filepath
            # to determine in which bin to find the target.
            digest_object = hashlib.new(name="sha256")
            digest_object.update(target_filepath.encode("utf-8"))
            target_filepath_hash = digest_object.hexdigest()

            for path_hash_prefix in self.path_hash_prefixes:
                if target_filepath_hash.startswith(path_hash_prefix):
                    return True

        elif self.paths is not None:
            for pathpattern in self.paths:
                # A delegated role path may be an explicit path or glob
                # pattern (Unix shell-style wildcards).
                if self._is_target_in_pathpattern(target_filepath, pathpattern):
                    return True

        return False


class SuccinctRoles(Role):
    """Succinctly defines a hash bin delegation graph.

    A ``SuccinctRoles`` object describes a delegation graph that covers all
    targets, distributing them uniformly over the delegated roles (i.e. bins)
    in the graph.

    The total number of bins is 2 to the power of the passed ``bit_length``.

    Bin names are the concatenation of the passed ``name_prefix`` and a
    zero-padded hex representation of the bin index separated by a hyphen.

    The passed ``keyids`` and ``threshold`` is used for each bin, and each bin
    is 'terminating'.

    For details: https://github.com/theupdateframework/taps/blob/master/tap15.md

    Args:
        keyids: Signing key identifiers for any bin metadata.
        threshold: Number of keys required to sign any bin metadata.
        bit_length: Number of bits between 1 and 32.
        name_prefix: Prefix of all bin names.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API.

    Raises:
            ValueError, TypeError, AttributeError: Invalid arguments.
    """

    def __init__(
        self,
        keyids: list[str],
        threshold: int,
        bit_length: int,
        name_prefix: str,
        unrecognized_fields: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(keyids, threshold, unrecognized_fields)

        if bit_length <= 0 or bit_length > 32:
            raise ValueError("bit_length must be between 1 and 32")
        if not isinstance(name_prefix, str):
            raise ValueError("name_prefix must be a string")

        self.bit_length = bit_length
        self.name_prefix = name_prefix

        # Calculate the suffix_len value based on the total number of bins in
        # hex. If bit_length = 10 then number_of_bins = 1024 or bin names will
        # have a suffix between "000" and "3ff" in hex and suffix_len will be 3
        # meaning the third bin will have a suffix of "003".
        self.number_of_bins = 2**bit_length
        # suffix_len is calculated based on "number_of_bins - 1" as the name
        # of the last bin contains the number "number_of_bins -1" as a suffix.
        self.suffix_len = len(f"{self.number_of_bins - 1:x}")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SuccinctRoles):
            return False

        return (
            super().__eq__(other)
            and self.bit_length == other.bit_length
            and self.name_prefix == other.name_prefix
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.bit_length, self.name_prefix))

    @classmethod
    def from_dict(cls, role_dict: dict[str, Any]) -> SuccinctRoles:
        """Create ``SuccinctRoles`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, AttributeError, TypeError: Invalid arguments.
        """
        keyids = role_dict.pop("keyids")
        threshold = role_dict.pop("threshold")
        bit_length = role_dict.pop("bit_length")
        name_prefix = role_dict.pop("name_prefix")
        # All fields left in the role_dict are unrecognized.
        return cls(keyids, threshold, bit_length, name_prefix, role_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        base_role_dict = super().to_dict()
        return {
            "bit_length": self.bit_length,
            "name_prefix": self.name_prefix,
            **base_role_dict,
        }

    def get_role_for_target(self, target_filepath: str) -> str:
        """Calculate the name of the delegated role responsible for
        ``target_filepath``.

        The target at path ``target_filepath`` is assigned to a bin by casting
        the left-most ``bit_length`` of bits of the file path hash digest to
        int, using it as bin index between 0 and ``2**bit_length - 1``.

        Args:
            target_filepath: URL path to a target file, relative to a base
                targets URL.
        """
        hasher = hashlib.new(name="sha256")
        hasher.update(target_filepath.encode("utf-8"))

        # We can't ever need more than 4 bytes (32 bits).
        hash_bytes = hasher.digest()[:4]
        # Right shift hash bytes, so that we only have the leftmost
        # bit_length bits that we care about.
        shift_value = 32 - self.bit_length
        bin_number = int.from_bytes(hash_bytes, byteorder="big") >> shift_value
        # Add zero padding if necessary and cast to hex the suffix.
        suffix = f"{bin_number:0{self.suffix_len}x}"
        return f"{self.name_prefix}-{suffix}"

    def get_roles(self) -> Iterator[str]:
        """Yield the names of all different delegated roles one by one."""
        for i in range(self.number_of_bins):
            suffix = f"{i:0{self.suffix_len}x}"
            yield f"{self.name_prefix}-{suffix}"

    def is_delegated_role(self, role_name: str) -> bool:
        """Determine whether the given ``role_name`` is in one of
        the delegated roles that ``SuccinctRoles`` represents.

        Args:
            role_name: The name of the role to check against.
        """
        desired_prefix = self.name_prefix + "-"

        if not role_name.startswith(desired_prefix):
            return False

        suffix = role_name[len(desired_prefix) :]
        if len(suffix) != self.suffix_len:
            return False

        try:
            # make sure suffix is hex value
            num = int(suffix, 16)
        except ValueError:
            return False

        return 0 <= num < self.number_of_bins


class Delegations:
    """A container object storing information about all delegations.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        keys: Dictionary of keyids to Keys. Defines the keys used in ``roles``.
        roles: Ordered dictionary of role names to DelegatedRoles instances. It
            defines which keys are required to sign the metadata for a specific
            role. The roles order also defines the order that role delegations
            are considered during target searches.
        succinct_roles: Contains succinct information about hash bin
            delegations. Note that succinct roles is not a TUF specification
            feature yet and setting `succinct_roles` to a value makes the
            resulting metadata non-compliant. The metadata will not be accepted
            as valid by specification compliant clients such as those built with
            python-tuf <= 1.1.0. For more information see: https://github.com/theupdateframework/taps/blob/master/tap15.md
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Exactly one of ``roles`` and ``succinct_roles`` must be set.

    Raises:
        ValueError: Invalid arguments.
    """

    def __init__(
        self,
        keys: dict[str, Key],
        roles: dict[str, DelegatedRole] | None = None,
        succinct_roles: SuccinctRoles | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        self.keys = keys
        if sum(1 for v in [roles, succinct_roles] if v is not None) != 1:
            raise ValueError("One of roles and succinct_roles must be set")

        if roles is not None:
            for role in roles:
                if not role or role in TOP_LEVEL_ROLE_NAMES:
                    raise ValueError(
                        "Delegated roles cannot be empty or use top-level "
                        "role names"
                    )

        self.roles = roles
        self.succinct_roles = succinct_roles
        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Delegations):
            return False

        all_attributes_check = (
            self.keys == other.keys
            and self.roles == other.roles
            and self.succinct_roles == other.succinct_roles
            and self.unrecognized_fields == other.unrecognized_fields
        )

        if self.roles is not None and other.roles is not None:
            all_attributes_check = (
                all_attributes_check
                # Order of the delegated roles matters (see issue #1788).
                and list(self.roles.items()) == list(other.roles.items())
            )

        return all_attributes_check

    def __hash__(self) -> int:
        return hash(
            (
                self.keys,
                self.roles,
                self.succinct_roles,
                self.unrecognized_fields,
            )
        )

    @classmethod
    def from_dict(cls, delegations_dict: dict[str, Any]) -> Delegations:
        """Create ``Delegations`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, TypeError: Invalid arguments.
        """
        keys = delegations_dict.pop("keys")
        keys_res = {}
        for keyid, key_dict in keys.items():
            keys_res[keyid] = Key.from_dict(keyid, key_dict)
        roles = delegations_dict.pop("roles", None)
        roles_res: dict[str, DelegatedRole] | None = None

        if roles is not None:
            roles_res = {}
            for role_dict in roles:
                new_role = DelegatedRole.from_dict(role_dict)
                if new_role.name in roles_res:
                    raise ValueError(f"Duplicate role {new_role.name}")
                roles_res[new_role.name] = new_role

        succinct_roles_dict = delegations_dict.pop("succinct_roles", None)
        succinct_roles_info = None
        if succinct_roles_dict is not None:
            succinct_roles_info = SuccinctRoles.from_dict(succinct_roles_dict)

        # All fields left in the delegations_dict are unrecognized.
        return cls(keys_res, roles_res, succinct_roles_info, delegations_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        keys = {keyid: key.to_dict() for keyid, key in self.keys.items()}
        res_dict: dict[str, Any] = {
            "keys": keys,
            **self.unrecognized_fields,
        }
        if self.roles is not None:
            roles = [role_obj.to_dict() for role_obj in self.roles.values()]
            res_dict["roles"] = roles
        elif self.succinct_roles is not None:
            res_dict["succinct_roles"] = self.succinct_roles.to_dict()

        return res_dict

    def get_roles_for_target(
        self, target_filepath: str
    ) -> Iterator[tuple[str, bool]]:
        """Given ``target_filepath`` get names and terminating status of all
        delegated roles who are responsible for it.

        Args:
            target_filepath: URL path to a target file, relative to a base
                targets URL.
        """
        if self.roles is not None:
            for role in self.roles.values():
                if role.is_delegated_path(target_filepath):
                    yield role.name, role.terminating

        elif self.succinct_roles is not None:
            # We consider all succinct_roles as terminating.
            # For more information read TAP 15.
            yield self.succinct_roles.get_role_for_target(target_filepath), True


class TargetFile(BaseFile):
    """A container with information about a particular target file.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        length: Length of the target file in bytes.
        hashes: Dictionary of hash algorithm names to hashes of the target
            file content.
        path: URL path to a target file, relative to a base targets URL.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError, TypeError: Invalid arguments.
    """

    def __init__(
        self,
        length: int,
        hashes: dict[str, str],
        path: str,
        unrecognized_fields: dict[str, Any] | None = None,
    ):
        self._validate_length(length)
        self._validate_hashes(hashes)

        self.length = length
        self.hashes = hashes
        self.path = path
        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    @property
    def custom(self) -> Any:  # noqa: ANN401
        """Get implementation specific data related to the target.

        python-tuf does not use or validate this data.
        """
        return self.unrecognized_fields.get("custom")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TargetFile):
            return False

        return (
            self.length == other.length
            and self.hashes == other.hashes
            and self.path == other.path
            and self.unrecognized_fields == other.unrecognized_fields
        )

    def __hash__(self) -> int:
        return hash(
            (self.length, self.hashes, self.path, self.unrecognized_fields)
        )

    @classmethod
    def from_dict(cls, target_dict: dict[str, Any], path: str) -> TargetFile:
        """Create ``TargetFile`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, TypeError: Invalid arguments.
        """
        length = target_dict.pop("length")
        hashes = target_dict.pop("hashes")

        # All fields left in the target_dict are unrecognized.
        return cls(length, hashes, path, target_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the JSON-serializable dictionary representation of self."""
        return {
            "length": self.length,
            "hashes": self.hashes,
            **self.unrecognized_fields,
        }

    @classmethod
    def from_file(
        cls,
        target_file_path: str,
        local_path: str,
        hash_algorithms: list[str] | None = None,
    ) -> TargetFile:
        """Create ``TargetFile`` object from a file.

        Args:
            target_file_path: URL path to a target file, relative to a base
                targets URL.
            local_path: Local path to target file content.
            hash_algorithms: Hash algorithms to calculate hashes with. If not
                specified, "sha256" is used.

        Raises:
            FileNotFoundError: The file doesn't exist.
            ValueError: The hash algorithms list contains an unsupported
                algorithm.
        """
        with open(local_path, "rb") as file:
            return cls.from_data(target_file_path, file, hash_algorithms)

    @classmethod
    def from_data(
        cls,
        target_file_path: str,
        data: bytes | IO[bytes],
        hash_algorithms: list[str] | None = None,
    ) -> TargetFile:
        """Create ``TargetFile`` object from bytes.

        Args:
            target_file_path: URL path to a target file, relative to a base
                targets URL.
            data: Target file content.
            hash_algorithms: Hash algorithms to create the hashes with. If not
                specified, "sha256" is used.

        Raises:
            ValueError: The hash algorithms list contains an unsupported
                algorithm.
        """
        length, hashes = cls._get_length_and_hashes(data, hash_algorithms)
        return cls(length, hashes, target_file_path)

    def verify_length_and_hashes(self, data: bytes | IO[bytes]) -> None:
        """Verify that length and hashes of ``data`` match expected values.

        Args:
            data: Target file object or its content in bytes.

        Raises:
            LengthOrHashMismatchError: Calculated length or hashes do not
                match expected values or hash algorithm is not supported.
        """
        self._verify_length(data, self.length)
        self._verify_hashes(data, self.hashes)

    def get_prefixed_paths(self) -> list[str]:
        """
        Return hash-prefixed URL path fragments for the target file path.
        """
        paths = []
        parent, sep, name = self.path.rpartition("/")
        for hash_value in self.hashes.values():
            paths.append(f"{parent}{sep}{hash_value}.{name}")

        return paths


class Targets(Signed, _DelegatorMixin):
    """A container for the signed part of targets metadata.

    Targets contains verifying information about target files and also
    delegates responsibility to other Targets roles.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        version: Metadata version number. Default is 1.
        spec_version: Supported TUF specification version. Default is the
            version currently supported by the library.
        expires: Metadata expiry date. Default is current date and time.
        targets: Dictionary of target filenames to TargetFiles. Default is an
            empty dictionary.
        delegations: Defines how this Targets delegates responsibility to other
            Targets Metadata files. Default is None.
        unrecognized_fields: Dictionary of all attributes that are not managed
            by TUF Metadata API

    Raises:
        ValueError: Invalid arguments.
    """

    type = _TARGETS

    def __init__(
        self,
        version: int | None = None,
        spec_version: str | None = None,
        expires: datetime | None = None,
        targets: dict[str, TargetFile] | None = None,
        delegations: Delegations | None = None,
        unrecognized_fields: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(version, spec_version, expires, unrecognized_fields)
        self.targets = targets if targets is not None else {}
        self.delegations = delegations

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Targets):
            return False

        return (
            super().__eq__(other)
            and self.targets == other.targets
            and self.delegations == other.delegations
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.targets, self.delegations))

    @classmethod
    def from_dict(cls, signed_dict: dict[str, Any]) -> Targets:
        """Create ``Targets`` object from its json/dict representation.

        Raises:
            ValueError, KeyError, TypeError: Invalid arguments.
        """
        common_args = cls._common_fields_from_dict(signed_dict)
        targets = signed_dict.pop(_TARGETS)
        try:
            delegations_dict = signed_dict.pop("delegations")
        except KeyError:
            delegations = None
        else:
            delegations = Delegations.from_dict(delegations_dict)
        res_targets = {}
        for target_path, target_info in targets.items():
            res_targets[target_path] = TargetFile.from_dict(
                target_info, target_path
            )
        # All fields left in the targets_dict are unrecognized.
        return cls(*common_args, res_targets, delegations, signed_dict)

    def to_dict(self) -> dict[str, Any]:
        """Return the dict representation of self."""
        targets_dict = self._common_fields_to_dict()
        targets = {}
        for target_path, target_file_obj in self.targets.items():
            targets[target_path] = target_file_obj.to_dict()
        targets_dict[_TARGETS] = targets
        if self.delegations is not None:
            targets_dict["delegations"] = self.delegations.to_dict()
        return targets_dict

    def add_key(self, key: Key, role: str | None = None) -> None:
        """Add new signing key for delegated role ``role``.

        If succinct_roles is used then the ``role`` argument is not required.

        Args:
            key: Signing key to be added for ``role``.
            role: Name of the role, for which ``key`` is added.

        Raises:
            ValueError: If the argument order is wrong or if there are no
                delegated roles or if ``role`` is not delegated by this Target.
        """
        # Verify that our users are not using the old argument order.
        if isinstance(role, Key):
            raise ValueError("Role must be a string, not a Key instance")

        if self.delegations is None:
            raise ValueError(f"Delegated role {role} doesn't exist")

        if self.delegations.roles is not None:
            if role not in self.delegations.roles:
                raise ValueError(f"Delegated role {role} doesn't exist")
            if key.keyid not in self.delegations.roles[role].keyids:
                self.delegations.roles[role].keyids.append(key.keyid)

        elif self.delegations.succinct_roles is not None:
            if key.keyid not in self.delegations.succinct_roles.keyids:
                self.delegations.succinct_roles.keyids.append(key.keyid)

        self.delegations.keys[key.keyid] = key

    def revoke_key(self, keyid: str, role: str | None = None) -> None:
        """Revokes key from delegated role ``role`` and updates the delegations
        key store.

        If succinct_roles is used then the ``role`` argument is not required.

        Args:
            keyid: Identifier of the key to be removed for ``role``.
            role: Name of the role, for which a signing key is removed.

        Raises:
            ValueError: If there are no delegated roles or if ``role`` is not
                delegated by this ``Target`` or if key is not used by ``role``
                or if key with id ``keyid`` is not used by succinct roles.
        """
        if self.delegations is None:
            raise ValueError(f"Delegated role {role} doesn't exist")

        if self.delegations.roles is not None:
            if role not in self.delegations.roles:
                raise ValueError(f"Delegated role {role} doesn't exist")
            if keyid not in self.delegations.roles[role].keyids:
                raise ValueError(f"Key with id {keyid} is not used by {role}")

            self.delegations.roles[role].keyids.remove(keyid)
            for keyinfo in self.delegations.roles.values():
                if keyid in keyinfo.keyids:
                    return

        elif self.delegations.succinct_roles is not None:
            if keyid not in self.delegations.succinct_roles.keyids:
                raise ValueError(
                    f"Key with id {keyid} is not used by succinct_roles"
                )

            self.delegations.succinct_roles.keyids.remove(keyid)

        del self.delegations.keys[keyid]

    def get_delegated_role(self, delegated_role: str) -> Role:
        """Return the role object for the given delegated role.

        Raises ValueError if delegated_role is not actually delegated.
        """
        if self.delegations is None:
            raise ValueError("No delegations found")

        role: Role | None = None
        if self.delegations.roles is not None:
            role = self.delegations.roles.get(delegated_role)
        elif self.delegations.succinct_roles is not None:
            succinct = self.delegations.succinct_roles
            if succinct.is_delegated_role(delegated_role):
                role = succinct

        if not role:
            raise ValueError(f"Delegated role {delegated_role} not found")

        return role

    def get_key(self, keyid: str) -> Key:
        if self.delegations is None:
            raise ValueError("No delegations found")
        if keyid not in self.delegations.keys:
            raise ValueError(f"Key {keyid} not found")

        return self.delegations.keys[keyid]



# Copyright 2021, New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""Test updating delegated targets roles and searching for
target files with various delegation graphs"""

from __future__ import annotations

import os
import sys
import tempfile
import unittest
from dataclasses import astuple, dataclass, field
from typing import TYPE_CHECKING

from tests import utils
from tests.repository_simulator import RepositorySimulator
from tuf.api.exceptions import UnsignedMetadataError
from tuf.api.metadata import (
    SPECIFICATION_VERSION,
    TOP_LEVEL_ROLE_NAMES,
    DelegatedRole,
    Targets,
)
from tuf.ngclient import Updater

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass
class TestDelegation:
    delegator: str
    rolename: str
    keyids: list[str] = field(default_factory=list)
    threshold: int = 1
    terminating: bool = False
    paths: list[str] | None = field(default_factory=lambda: ["*"])
    path_hash_prefixes: list[str] | None = None


@dataclass
class TestTarget:
    rolename: str
    content: bytes
    targetpath: str


@dataclass
class DelegationsTestCase:
    """A delegations graph as lists of delegations and target files
    and the expected order of traversal as a list of role names."""

    delegations: list[TestDelegation]
    target_files: list[TestTarget] = field(default_factory=list)
    visited_order: list[str] = field(default_factory=list)


@dataclass
class TargetTestCase:
    targetpath: str
    found: bool
    visited_order: list[str] = field(default_factory=list)


class TestDelegations(unittest.TestCase):
    """Base class for delegation tests"""

    # set dump_dir to trigger repository state dumps
    dump_dir: str | None = None

    def setUp(self) -> None:
        self.subtest_count = 0
        self.temp_dir = tempfile.TemporaryDirectory()
        self.metadata_dir = os.path.join(self.temp_dir.name, "metadata")
        self.targets_dir = os.path.join(self.temp_dir.name, "targets")
        os.mkdir(self.metadata_dir)
        os.mkdir(self.targets_dir)
        self.sim: RepositorySimulator

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def setup_subtest(self) -> None:
        self.subtest_count += 1
        if self.dump_dir is not None:
            # create subtest dumpdir
            name = f"{self.id().split('.')[-1]}-{self.subtest_count}"
            self.sim.dump_dir = os.path.join(self.dump_dir, name)
            os.mkdir(self.sim.dump_dir)
            # dump the repo simulator metadata
            self.sim.write()

    def teardown_subtest(self) -> None:
        utils.cleanup_metadata_dir(self.metadata_dir)

    def _init_repo(self, test_case: DelegationsTestCase) -> None:
        """Create a new RepositorySimulator instance and
        populate it with delegations and target files"""

        self.sim = RepositorySimulator()
        spec_version = ".".join(SPECIFICATION_VERSION)
        for d in test_case.delegations:
            if d.rolename in self.sim.md_delegates:
                targets = self.sim.md_delegates[d.rolename].signed
            else:
                targets = Targets(
                    1, spec_version, self.sim.safe_expiry, {}, None
                )
            # unpack 'd' but skip "delegator"
            role = DelegatedRole(*astuple(d)[1:])
            self.sim.add_delegation(d.delegator, role, targets)

        for target in test_case.target_files:
            self.sim.add_target(*astuple(target))

        if test_case.target_files:
            self.sim.targets.version += 1
        self.sim.update_snapshot()

    def _init_updater(self) -> Updater:
        """Create a new Updater instance"""
        return Updater(
            self.metadata_dir,
            "https://example.com/metadata/",
            self.targets_dir,
            "https://example.com/targets/",
            self.sim,
            bootstrap=self.sim.signed_roots[0],
        )

    def _assert_files_exist(self, roles: Iterable[str]) -> None:
        """Assert that local metadata files match 'roles'"""
        expected_files = [f"{role}.json" for role in roles]
        found_files = [
            e.name for e in os.scandir(self.metadata_dir) if e.is_file()
        ]

        self.assertListEqual(sorted(found_files), sorted(expected_files))


class TestDelegationsGraphs(TestDelegations):
    """Test creating delegations graphs with different complexity
    and successfully updating the delegated roles metadata"""

    graphs = {
        "basic delegation": DelegationsTestCase(
            delegations=[TestDelegation("targets", "A")],
            visited_order=["A"],
        ),
        "single level delegations": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
            ],
            visited_order=["A", "B"],
        ),
        "two-level delegations": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("B", "C"),
            ],
            visited_order=["A", "B", "C"],
        ),
        "two-level test DFS order of traversal": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C"),
                TestDelegation("A", "D"),
            ],
            visited_order=["A", "C", "D", "B"],
        ),
        "three-level delegation test DFS order of traversal": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C"),
                TestDelegation("C", "D"),
            ],
            visited_order=["A", "C", "D", "B"],
        ),
        "two-level terminating ignores all but role's descendants": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C", terminating=True),
                TestDelegation("A", "D"),
            ],
            visited_order=["A", "C"],
        ),
        "three-level terminating ignores all but role's descendants": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C", terminating=True),
                TestDelegation("C", "D"),
            ],
            visited_order=["A", "C", "D"],
        ),
        "two-level ignores all branches not matching 'paths'": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A", paths=["*.py"]),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C"),
            ],
            visited_order=["B"],
        ),
        "three-level ignores all branches not matching 'paths'": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("A", "C", paths=["*.py"]),
                TestDelegation("C", "D"),
            ],
            visited_order=["A", "B"],
        ),
        "cyclic graph": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("B", "C"),
                TestDelegation("C", "D"),
                TestDelegation("D", "B"),
            ],
            visited_order=["A", "B", "C", "D"],
        ),
        "two roles delegating to a third": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("B", "C"),
                TestDelegation("A", "C"),
            ],
            # Under all same conditions, 'C' is reached through 'A' first"
            visited_order=["A", "C", "B"],
        ),
        "two roles delegating to a third different 'paths'": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("B", "C"),
                TestDelegation("A", "C", paths=["*.py"]),
            ],
            # 'C' is reached through 'B' since 'A' does not delegate a matching pattern"
            visited_order=["A", "B", "C"],
        ),
        "max number of delegations": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "A"),
                TestDelegation("targets", "B"),
                TestDelegation("targets", "C"),
                TestDelegation("C", "D"),
                TestDelegation("C", "E"),
            ],
            # "E" is skipped, max_delegations is 4
            visited_order=["A", "B", "C", "D"],
        ),
    }

    @utils.run_sub_tests_with_dataset(graphs)
    def test_graph_traversal(self, test_data: DelegationsTestCase) -> None:
        """Test that delegated roles are traversed in the order of appearance
        in the delegator's metadata, using pre-order depth-first search"""

        try:
            exp_files = [*TOP_LEVEL_ROLE_NAMES, *test_data.visited_order]
            exp_calls = [(role, 1) for role in test_data.visited_order]

            self._init_repo(test_data)
            self.setup_subtest()

            updater = self._init_updater()
            # restrict the max number of delegations to simplify the test
            updater.config.max_delegations = 4
            # Call explicitly refresh to simplify the expected_calls list
            updater.refresh()
            self.sim.fetch_tracker.metadata.clear()
            # Check that metadata dir contains only top-level roles
            self._assert_files_exist(TOP_LEVEL_ROLE_NAMES)

            # Looking for a non-existing targetpath forces updater
            # to visit all possible delegated roles
            targetfile = updater.get_targetinfo("missingpath")
            self.assertIsNone(targetfile)
            # Check that the delegated roles were visited in the expected
            # order and the corresponding metadata files were persisted
            self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)
            self._assert_files_exist(exp_files)
        finally:
            self.teardown_subtest()

    invalid_metadata = {
        "unsigned delegated role": DelegationsTestCase(
            delegations=[
                TestDelegation("targets", "invalid"),
                TestDelegation("targets", "B"),
                TestDelegation("invalid", "C"),
            ],
            # The traversal stops after visiting an invalid role
            visited_order=["invalid"],
        )
    }

    @utils.run_sub_tests_with_dataset(invalid_metadata)
    def test_invalid_metadata(self, test_data: DelegationsTestCase) -> None:
        try:
            self._init_repo(test_data)
            # The invalid role is the last visited
            invalid_role = test_data.visited_order[-1]
            self.sim.signers[invalid_role].clear()

            self.setup_subtest()
            # The invalid role metadata must not be persisted
            exp_files = [*TOP_LEVEL_ROLE_NAMES, *test_data.visited_order[:-1]]
            exp_calls = [(role, 1) for role in test_data.visited_order]

            updater = self._init_updater()
            # Call explicitly refresh to simplify the expected_calls list
            updater.refresh()
            self.sim.fetch_tracker.metadata.clear()

            with self.assertRaises(UnsignedMetadataError):
                updater.get_targetinfo("missingpath")
            # Check that there were no visited roles after the invalid one
            # and only the valid metadata files were persisted
            self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)
            self._assert_files_exist(exp_files)
        finally:
            self.teardown_subtest()

    def test_safely_encoded_rolenames(self) -> None:
        """Test that delegated roles names are safely encoded in the filenames
        and URLs.
        """

        roles_to_filenames = {
            "../a": "..%2Fa.json",
            ".": "..json",
            "/": "%2F.json",
            "ö": "%C3%B6.json",
        }

        delegations = []
        for rolename in roles_to_filenames:
            delegations.append(TestDelegation("targets", rolename))

        delegated_rolenames = DelegationsTestCase(delegations)
        self._init_repo(delegated_rolenames)
        updater = self._init_updater()
        updater.refresh()

        # trigger updater to fetch the delegated metadata
        self.sim.fetch_tracker.metadata.clear()
        updater.get_targetinfo("anything")

        # assert that local delegated metadata filenames are expected
        local_metadata = os.listdir(self.metadata_dir)
        for fname in roles_to_filenames.values():
            self.assertTrue(fname in local_metadata)

        # assert that requested URLs are quoted without extension
        exp_calls = [(quoted[:-5], 1) for quoted in roles_to_filenames.values()]
        self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)

    hash_bins_graph = {
        "delegations": DelegationsTestCase(
            delegations=[
                TestDelegation(
                    "targets",
                    "role1",
                    paths=None,
                    path_hash_prefixes=["8", "9", "a", "b"],
                ),
                TestDelegation(
                    "targets",
                    "role2",
                    paths=None,
                    path_hash_prefixes=["0", "1", "2", "3"],
                ),
                TestDelegation(
                    "targets",
                    "role3",
                    paths=None,
                    path_hash_prefixes=["c", "d", "e", "f"],
                ),
            ],
            visited_order=["role1", "role2", "role3"],
        ),
    }

    @utils.run_sub_tests_with_dataset(hash_bins_graph)
    def test_hash_bins_graph_traversal(
        self, test_data: DelegationsTestCase
    ) -> None:
        """Test that delegated roles are traversed in the order of appearance
        in the delegator's metadata, using pre-order depth-first search and that
        they correctly refer to the corresponding hash bin prefixes"""

        try:
            exp_files = [*TOP_LEVEL_ROLE_NAMES, *test_data.visited_order]
            exp_calls = [(role, 1) for role in test_data.visited_order]

            self._init_repo(test_data)
            self.setup_subtest()

            updater = self._init_updater()
            # Call explicitly refresh to simplify the expected_calls list
            updater.refresh()
            self.sim.fetch_tracker.metadata.clear()
            # Check that metadata dir contains only top-level roles
            self._assert_files_exist(TOP_LEVEL_ROLE_NAMES)

            # Looking for a non-existing targetpath forces updater
            # to visit a correspondning delegated role
            targetfile = updater.get_targetinfo("missingpath")
            self.assertIsNone(targetfile)
            targetfile = updater.get_targetinfo("othermissingpath")
            self.assertIsNone(targetfile)
            targetfile = updater.get_targetinfo("thirdmissingpath")
            self.assertIsNone(targetfile)
            # Check that the delegated roles were visited in the expected
            # order and the corresponding metadata files were persisted
            self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)
            self._assert_files_exist(exp_files)
        finally:
            self.teardown_subtest()

    @dataclass
    class SuccinctRolesTestCase:
        bit_length: int
        target_path: str
        expected_target_bin: str

    # By setting the bit_length the total number of bins is 2^bit_length.
    # In each test case target_path is a path to a random target we want to
    # fetch and expected_target_bin is the bin we are expecting to visit.
    succinct_bins_graph = {
        "bin amount = 2, target bin index 0": SuccinctRolesTestCase(
            bit_length=1,
            target_path="boo",
            expected_target_bin="bin-0",
        ),
        "bin amount = 2, target bin index 1": SuccinctRolesTestCase(
            bit_length=1,
            target_path="too",
            expected_target_bin="bin-1",
        ),
        "bin amount = 4, target bin index 0": SuccinctRolesTestCase(
            bit_length=2,
            target_path="foo",
            expected_target_bin="bin-0",
        ),
        "bin amount = 4, target bin index 1": SuccinctRolesTestCase(
            bit_length=2,
            target_path="doo",
            expected_target_bin="bin-1",
        ),
        "bin amount = 4, target bin index 2": SuccinctRolesTestCase(
            bit_length=2,
            target_path="too",
            expected_target_bin="bin-2",
        ),
        "bin amount = 4, target bin index 3": SuccinctRolesTestCase(
            bit_length=2,
            target_path="bar",
            expected_target_bin="bin-3",
        ),
        "bin amount = 256, target bin index fc": SuccinctRolesTestCase(
            bit_length=8,
            target_path="bar",
            expected_target_bin="bin-fc",
        ),
    }

    @utils.run_sub_tests_with_dataset(succinct_bins_graph)
    def test_succinct_roles_graph_traversal(
        self, test_data: SuccinctRolesTestCase
    ) -> None:
        # Test traversing the delegation tree when succinct roles is used. For a
        # successful traversal all top level metadata files plus the expected
        # bin should exist locally and only one bin must be downloaded.

        try:
            exp_files = [*TOP_LEVEL_ROLE_NAMES, test_data.expected_target_bin]
            exp_calls = [(test_data.expected_target_bin, 1)]

            self.sim = RepositorySimulator()
            self.sim.add_succinct_roles("targets", test_data.bit_length, "bin")
            self.sim.update_snapshot()

            self.setup_subtest()

            updater = self._init_updater()
            # Call explicitly refresh to simplify the expected_calls list.
            updater.refresh()
            self.sim.fetch_tracker.metadata.clear()
            # Check that metadata dir contains only top-level roles
            self._assert_files_exist(TOP_LEVEL_ROLE_NAMES)

            # Looking for a non-existing targetpath forces updater
            # to visit a corresponding delegated role.
            targetfile = updater.get_targetinfo(test_data.target_path)
            self.assertIsNone(targetfile)

            # Check that the delegated roles were visited in the expected
            # order and the corresponding metadata files were persisted.
            self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)
            self._assert_files_exist(exp_files)

        finally:
            self.teardown_subtest()


class TestTargetFileSearch(TestDelegations):
    r"""
    Create a single repository with the following delegations:

              targets
    *.doc, *md / \ release/*/*
              A   B
     release/x/* / \ release/y/*.zip
                C   D

    Test that Updater successfully finds the target files metadata,
    traversing the delegations as expected.
    """

    delegations_tree = DelegationsTestCase(
        delegations=[
            TestDelegation("targets", "A", paths=["*.doc", "*.md"]),
            TestDelegation("targets", "B", paths=["releases/*/*"]),
            TestDelegation("B", "C", paths=["releases/x/*"]),
            TestDelegation("B", "D", paths=["releases/y/*.zip"]),
        ],
        target_files=[
            TestTarget("targets", b"targetfile content", "targetfile"),
            TestTarget("A", b"README by A", "README.md"),
            TestTarget("C", b"x release by C", "releases/x/x_v1"),
            TestTarget("D", b"y release by D", "releases/y/y_v1.zip"),
            TestTarget("D", b"z release by D", "releases/z/z_v1.zip"),
        ],
    )

    def setUp(self) -> None:
        super().setUp()
        self._init_repo(self.delegations_tree)

    # fmt: off
    targets = {
        "no delegations":
            TargetTestCase("targetfile", True, []),
        "targetpath matches wildcard":
            TargetTestCase("README.md", True, ["A"]),
        "targetpath with separators x":
            TargetTestCase("releases/x/x_v1", True, ["B", "C"]),
        "targetpath with separators y":
            TargetTestCase("releases/y/y_v1.zip", True, ["B", "D"]),
        "targetpath is not delegated by all roles in the chain":
            TargetTestCase("releases/z/z_v1.zip", False, ["B"]),
    }
    # fmt: on

    @utils.run_sub_tests_with_dataset(targets)
    def test_targetfile_search(self, test_data: TargetTestCase) -> None:
        try:
            self.setup_subtest()
            exp_files = [*TOP_LEVEL_ROLE_NAMES, *test_data.visited_order]
            exp_calls = [(role, 1) for role in test_data.visited_order]
            exp_target = self.sim.target_files[test_data.targetpath].target_file

            updater = self._init_updater()
            # Call explicitly refresh to simplify the expected_calls list
            updater.refresh()
            self.sim.fetch_tracker.metadata.clear()
            target = updater.get_targetinfo(test_data.targetpath)
            if target is not None:
                # Confirm that the expected TargetFile is found
                self.assertTrue(test_data.found)
                self.assertDictEqual(target.to_dict(), exp_target.to_dict())
            else:
                self.assertFalse(test_data.found)
            # Check that the delegated roles were visited in the expected
            # order and the corresponding metadata files were persisted
            self.assertListEqual(self.sim.fetch_tracker.metadata, exp_calls)
            self._assert_files_exist(exp_files)
        finally:
            self.teardown_subtest()


if __name__ == "__main__":
    if "--dump" in sys.argv:
        TestDelegations.dump_dir = tempfile.mkdtemp()
        print(f"Repository Simulator dumps in {TestDelegations.dump_dir}")
        sys.argv.remove("--dump")

    utils.configure_test_logging(sys.argv)
    unittest.main()



"""Script for generating new metadata files."""

# Copyright New York University and the TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

from __future__ import annotations

import os
import sys
from datetime import datetime, timezone

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from securesystemslib.signer import CryptoSigner, Signer, SSlibKey

from tests import utils
from tuf.api.metadata import Metadata, Root, Snapshot, Targets, Timestamp
from tuf.api.serialization.json import JSONSerializer

# Hardcode keys and expiry time to achieve reproducibility.
public_values: list[str] = [
    "b11d2ff132c033a657318c74c39526476c56de7556c776f11070842dbc4ac14c",
    "250f9ae3d1d3d5c419a73cfb4a470c01de1d5d3d61a3825416b5f5d6b88f4a30",
    "82380623abb9666d4bf274b1a02577469445a972e5650d270101faa5107b19c8",
    "0e6738fc1ac6fb4de680b4be99ecbcd99b030f3963f291277eef67bb9bd123e9",
]
private_values: list[bytes] = [
    bytes.fromhex(
        "510e5e04d7a364af850533856eacdf65d30cc0f8803ecd5fdc0acc56ca2aa91c"
    ),
    bytes.fromhex(
        "e6645b00312c8a257782e3e61e85bafda4317ad072c52251ef933d480c387abd"
    ),
    bytes.fromhex(
        "cd13dd2180334b24c19b32aaf27f7e375a614d7ba0777220d5c2290bb2f9b868"
    ),
    bytes.fromhex(
        "7e2e751145d1b22f6e40d4ba2aa47158207acfd3c003f1cbd5a08141dfc22a15"
    ),
]
keyids: list[str] = [
    "5822582e7072996c1eef1cec24b61115d364987faa486659fe3d3dce8dae2aba",
    "09d440e3725cec247dcb8703b646a87dd2a4d75343e8095c036c32795eefe3b9",
    "3458204ed467519c19a5316eb278b5608472a1bbf15850ebfb462d5315e4f86d",
    "2be5c21e3614f9f178fb49c4a34d0c18ffac30abd14ced917c60a52c8d8094b7",
]

signers: list[Signer] = []
for index in range(len(keyids)):
    key = SSlibKey(
        keyids[index],
        "ed25519",
        "ed25519",
        {"public": public_values[index]},
    )
    private_key = Ed25519PrivateKey.from_private_bytes(private_values[index])
    signers.append(CryptoSigner(private_key, key))

EXPIRY = datetime(2050, 1, 1, tzinfo=timezone.utc)

SERIALIZER = JSONSerializer()


def verify_generation(md: Metadata, path: str) -> None:
    """Verify that newly generated file equals the locally stored one.

    Args:
        md: Newly generated metadata object.
        path: Path to the locally stored metadata file.
    """
    with open(path, "rb") as f:
        static_md_bytes = f.read()
        md_bytes = md.to_bytes(SERIALIZER)
        if static_md_bytes != md_bytes:
            raise ValueError(
                f"Generated data != local data at {path}. Generate a new "
                "metadata with 'python generated_data/generate_md.py'"
            )


def generate_all_files(dump: bool = False) -> None:
    """Generate a new repository or verify that output has not changed.

    Args:
        dump: If True, new files are generated. If False, existing files
           are compared to generated files and an exception is raised if
           there are differences.
    """
    md_root = Metadata(Root(expires=EXPIRY))
    md_timestamp = Metadata(Timestamp(expires=EXPIRY))
    md_snapshot = Metadata(Snapshot(expires=EXPIRY))
    md_targets = Metadata(Targets(expires=EXPIRY))

    md_root.signed.add_key(signers[0].public_key, "root")
    md_root.signed.add_key(signers[1].public_key, "timestamp")
    md_root.signed.add_key(signers[2].public_key, "snapshot")
    md_root.signed.add_key(signers[3].public_key, "targets")

    for i, md in enumerate([md_root, md_timestamp, md_snapshot, md_targets]):
        assert isinstance(md, Metadata)
        md.sign(signers[i])
        path = os.path.join(
            utils.TESTS_DIR,
            "generated_data",
            "ed25519_metadata",
            f"{md.signed.type}_with_ed25519.json",
        )
        if dump:
            md.to_file(path, SERIALIZER)
        else:
            verify_generation(md, path)


if __name__ == "__main__":
    utils.configure_test_logging(sys.argv)
    # To generate a new set of metadata files this script is supposed to be run
    # from the "tests" folder.
    generate_all_files(dump=True)
